#!/usr/bin/env python3

############################################################################
# Copyright (C) 2019 by gempa GmbH                                         #
#                                                                          #
# All Rights Reserved.                                                     #
#                                                                          #
# NOTICE: All information contained herein is, and remains                 #
# the property of gempa GmbH and its suppliers, if any. The intellectual   #
# and technical concepts contained herein are proprietary to gempa GmbH    #
# and its suppliers.                                                       #
# Dissemination of this information or reproduction of this material       #
# is strictly forbidden unless prior written permission is obtained        #
# from gempa GmbH.                                                         #
#                                                                          #
#  Author: Enrico Ellguth, Ralph Henneberger                               #
#  Email: enrico.ellguth@gempa.de, ralix@gempa.de                          #
############################################################################

import os
import signal
import sys
import platform

import getopt
import logging

DEPS = [
    "configparser",
    "cryptography",  # hash sums
    "humanize",  # SI prefix
    "natsort",
    "python-dateutil",
    "pytz",
    "requests",  # package download
    "tqdm",  # progress bar
    "tzlocal",
    "numpy",
]

if sys.version_info < (3, 6):
    print(
        "gsm requires at least Python 3.6. You have "
        + str(platform.python_version() + "...")
    )
    sys.exit(1)


try:
    gsmBaseDir, _ = os.path.split(os.path.dirname(os.path.realpath(__file__)))
    sys.path.append(gsmBaseDir)

    from tqdm import tqdm

    from libs import utils, package, log

    from libs.cache import localcache, synccache
    from libs import config, makelinks, releaseupgrade, osupgrade, releasedowngrade

    from libs.utils import checkInput, getInput

    from libs.const import (
        PT,
        GSM_INFO_OBSOLETE_PACKAGES_URL,
        urljoin,
        BASE_PUBLIC_URL,
        SC,
        SC3,
        DEFAULT_DATE_FORMAT,
        NO_ARCH,
        ANY,
        DSC_EXTENSION,
        CRT_EXTENSION,
        LICENSE_TXT_EXTENSION,
        REMOVE_OK,
        MAX_LEN_PACKAGE_FULL_NAME,
    )

    from libs.package import Package

except ImportError as e:
    print(
        f"""{e}

You may install all dependencies locally using the following command

  python{sys.version_info.major} -m pip install --user {' '.join(DEPS)}

Note: Some packages may require additional packages, e.g., the Python development
package.
"""
    )
    sys.exit(0)

USAGE_INFO = """
gsm.py - gempa software manager is a command line tool for handling packages

Usage: python gsm.py [options] command [module1] [module2] ...

Commands:
    search                             Search the package list for a regex pattern
    install [module1 ...]              Install or update packages. To install a specific
                                       package run, e.g., install gds=2019.287. If no
                                       module is provided, all installed modules will be
                                       updated.
    localinstall [file1 ...]           Install packages from local file (e.g.,
                                       scanloc-2023.128-seiscomp-5-rhel-8-x86_64.tar.gz)
                                       Note: The package description file is expected in
                                       the current directory.
    install licenses [module1 ...]     Install or update licenses. If no module is
                                       provided, the licenses of all installed
                                       modules will be updated.
    reinstall [module1 ...]            Reinstall packages. If no module is provided,
                                       all installed modules will be reinstalled.
    do-release-upgrade                 Upgrade SeisComP release
    do-release-downgrade               Downgrade SeisComP release (only >= SeisComP 4)
    switch-os                          Change the operating system

    hold [module1 ...]                 Add packages to hold
    unhold [module1 ...]               Remove packages from hold

    show                               Show available modules
    show updates                       Show possible updates
    show [module]                      Show latest package for a single module
    show [module] changelog            Show module changelog (online)
    show releases                      Show available SeisComP releases
    show hold                          Show all hold packages
    show all [module1 ...]             Show all available packages 
    info                               Show installed modules
    info [module1 ...]                 Show detailed information about an installed module
    info short                         Show installed modules (only package names)
    setup                              Run initial setup (required when using for the first time)
    update                             Synchronize the package index from their source and show possible updates
    sync                               Synchronize the package index from their source
    migrate                            Updates config schema
    remove module1 [module2 ...]       Uninstall packages
    clean [dbcache] [packages] [all]   Remove saved data (DBCache, Packages, Packages + DBCache...)

Options:
        --url                          URL of the package directory, e.g., https://gempa-build/builds/deploy
        --noverify                     Do not verify server certificate
    -d, --downloadonly                 Do not install, just download
        --downloaddir=directory        Specifies an alternative directory to store packages
        --installdir=directory         Specifies an alternative directory to install packages
        --debug                        Use debug packages only
    -u, --user                         Username to use
    -p, --password                     Password to use
    -r, --release                      SeisComP release, e.g., jakarta-2016.161
        --arch                         System architecture, e.g., x86_64
        --showurls                     Display URLs
    -y, --assumeyes                    Assume that the answer to questions which would
                                       be asked is yes. Errors still require user interaction.
                                       Configuration Option: assumeyes
        --exclude=[module1,...]        Exclude packages from (re)install and remove
    -o, --offline                      Run gsm in offline modus
    -f, --force                        Disables the remote file time check when downloading
                                       files. With this option enabled file downloads are
                                       always performed.
    -h, --help                         Display this help message and exit

Examples:

* Workflow for installing the latest scanloc package from scratch:
  1. Set up the initial configuration (required when using for the first time)
  2. Synchronize the package index
  3. View the changelog of the package
  4. Install the latest scanloc package

  gsm setup (Run initial setup (required when using for the first time))
  gsm update
  seiscomp stop scanloc
  gsm show scanloc changelog
  gsm install scanloc
  seiscomp start scanloc

* Workflow for updating SeisComP, updating all other packages is therefore required:
  1. Set up the initial configuration
  2. Synchronize the package index
  3. Stop all modules
  4. Read the changelog of all packages, replace "<package>" by the name of the package.
  5. Install all packages
  6. Restart all packages

  gsm setup
  gsm update
  seiscomp stop
  gsm show
  gsm show <package> changelog
  gsm install
  seiscomp update-config   (updating the database tables may be required before,
                            read the SeisComP changelog)
  seiscomp start

"""

GSM_LOG_DIR = "log"
GSM_LOG_FILE = os.path.join(GSM_LOG_DIR, "gsm.log")


class App(object):
    def __init__(self, argc, argv):
        self.argv = argv
        self.argc = argc

        # the base dir is the parent directory of the gsm.py excecutable
        baseDir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        self.config = config.Config(baseDir)

        self.command = None
        self.localCache = None
        self.syncCache = None
        self.args = None
        self.excludePackages = None
        self.installRequested = False

    def handleCommand(self, command):
        if command == "help":
            return True

        if command == "setup":
            return self.config.setup()

        if not self.config.baseURL == BASE_PUBLIC_URL:
            if not self.config.user:
                print("No username specified")
                return False

            if not self.config.password:
                print("No password specified")
                return False

        if not self.config.release:
            print("No SeisComP release specified")
            return False

        if self.syncCache.isValid():
            if not os.path.exists(
                (os.path.join(self.config.syncDir, SC, self.config.archDir))
            ):
                print(
                    f"\nYour configured OS ({self.config.os} {self.config.os_version}"
                    f" ({self.config.arch})) is no longer supported!\n"
                )

            if not os.path.exists(
                (os.path.join(self.config.syncDir, SC, self.config.release))
            ):
                print(
                    f"\nYour configured SeisComP Release ({self.config.release}) is "
                    "no longer supported!\n"
                    "You can switch to a new release with ./gsm do-release-upgrade.\n"
                )

        # remove duplicates but keep order
        packages = []
        _ = [packages.append(x) for x in self.args if x not in packages]

        if not packages or len(packages) == 1 and packages[0].lower() == "all":
            packages = list(self.localCache.getIndex().keys())

        if packages and packages[0] == "licenses" and len(packages) == 1:
            packages.extend(list(self.localCache.getIndex().keys()))

        if command == "sync":
            if self.args:
                print("Sync command takes no arguments")
                return False
            self.config.fromUpdate = True
            self.setMaxlenPackageFullName(packages)
            return self.syncCache.updateIndex()

        if command == "update":
            if self.args:
                print("Update command takes no arguments")
                return False
            self.config.fromUpdate = True

            self.setMaxlenPackageFullName(packages, 7)

            self.syncCache.updateIndex()

            if self.syncCache.getIndex():
                ret, info = package.isNewReleaseAvailable(self.config)
                if ret:
                    print("")
                    print(info)

                ret, _ = package.updateCheck(
                    self.localCache, self.syncCache, self.config
                )
                if ret:
                    self.selfUpdate()

                ret, updatePackages = package.getPackagesUpdateInfo(
                    self.syncCache, self.localCache, self.config
                )
                if ret:
                    if not self.config.assumeYes:
                        if checkInput(
                            "\nDo you want to install packages which"
                            " can be updated?",
                            "y",
                            "Yn",
                        ):
                            self.installRequested = True
                        else:
                            return False
                    if updatePackages:
                        return self.install(updatePackages, False)

                    return self.install(packages, False)

                return True

            return False

        if self.syncCache.getIndex() is False:
            print(
                "No packages found for"
                f" {self.config.release}/{self.config.os}/"
                "{self.config.os_version}/{self.config.arch}"
            )
            return True

        if command == "search":
            if not packages:
                print("Command requires an argument")
                return False

            return self.syncCache.search(packages)

        if command == "clean":
            if not self.args or not len(self.args) == 1:
                print("Command requires only one argument: [dbcache] [packages] [all]")
                return False
            return utils.clean(self.args, self.config)

        if command == "show":
            if len(packages) == 1 and packages[0] == "updates":
                package.getPackagesUpdateInfo(
                    self.syncCache, self.localCache, self.config
                )
                return True

            if len(packages) == 1 and packages[0] == "releases":
                return package.showReleases(self.config)

            if len(packages) == 1 and packages[0] == "hold":
                return self.localCache.showHoldPackages()

            if len(packages) >= 2 and packages[0] == "all":
                packages.remove("all")
                return self.syncCache.infoPackages(packages, True)

            if len(packages) == 2 and packages[1] == "changelog":
                pkgs = self.syncCache.getPackages(packages[0])
                if not pkgs:
                    return False

                ret, showChangeLog = package.showChangeLog(
                    [pkgs[0].name],
                    self.config,
                    self.localCache,
                    self.syncCache,
                    self.config.offline,
                    False,
                )
                if ret and not showChangeLog:
                    return ret

            return self.syncCache.infoPackages(packages if self.args else None)

        if command == "info":
            if packages[0] == "short":
                return self.localCache.infoPackagesShort()

            return self.localCache.infoPackages(packages if self.args else None)

        # if command == "install-licenses":
        #     return self.installLicenses(packages)

        if command == "localinstall":
            return self.localInstall(packages)

        if command == "install":
            if packages[0] == "licenses":
                packages.remove("licenses")
                return self.installLicenses(packages)

            return self.install(packages, False)

        if command == "reinstall":
            installedPackages = []
            staticPackages = []
            packages = self.filterPackages(packages)

            for name in packages:
                pkg = self.localCache.getPackage(name)
                if not pkg:
                    log.logWarning(f"{name} is not installed ... ")
                    continue

                if pkg.type == PT.DATA:
                    staticPackages.append(name)
                    continue
                installedPackages.append(pkg.fullName() if pkg else name)
            if staticPackages:
                log.logInfo(
                    f"{', '.join(staticPackages)} cannot be reinstalled. This operation is not"
                    " supported for data packages."
                )
                log.logInfo(
                    "Please use ./gsm remove [module] and ./gsm install [module].\n"
                )
            return self.install(installedPackages, True)

        if command == "remove":
            return self.remove(packages)

        if command == "hold":
            for name in packages:
                pkg = self.localCache.getPackage(name)
                if not pkg:
                    log.logInfo(f"{name} is not installed ... ")
                else:
                    self.localCache.addHoldPackage(name)

            return True

        if command == "unhold":
            for name in packages:
                pkg = self.localCache.getPackage(name)
                if not pkg:
                    log.logInfo(f"{name} is not installed ... ")
                    return False

                self.localCache.removeHoldPackage(name)

            return True

        if command == "do-release-upgrade":
            upgrader = releaseupgrade.ReleaseUpgrade(self.config, self.localCache)
            return upgrader.run(self)

        if command == "do-release-downgrade":
            downgrader = releasedowngrade.ReleaseDowngrade(self.config, self.localCache)
            return downgrader.run(self)

        if command == "switch-os":
            upgrader = osupgrade.OSUpgrade(self.config, self.localCache)
            return upgrader.run(self)

        if command == "migrate":
            return self.config.makeMigration()

        print(f"Unknown command {command}")
        return False

    def init(self):
        # initial commandline parsing: check for help and setup command
        if not self.parseCommandLine(True):
            print(USAGE_INFO, file=sys.stderr)
            return False

        # print help message and exit
        if self.command == "help":
            print(USAGE_INFO)
            return True

        # configuration file exits
        if os.path.isfile(self.config.fileName):
            error = self.config.read()
            # error while reading configuration and command is not setup
            if error and self.command != "setup":
                print(error, file=sys.stderr)
                return False
            if self.command != "setup":
                if not self.config.checkInstallInfo():
                    return False
        # no configuration file found and command is not setup
        elif self.command != "setup":
            print(
                f"Configuration file '{self.config.fileName}' not found. Please run"
                " './gsm setup' to create an initial configuration"
            )
            return False

        # second stage command line parsing: command line options override
        # config file options
        if not self.parseCommandLine():
            print(USAGE_INFO, file=sys.stderr)
            return False

        if not utils.createDirectory(GSM_LOG_DIR):
            return False

        logging.basicConfig(
            filename=GSM_LOG_FILE,
            format="%(asctime)s - %(message)s",
            datefmt=DEFAULT_DATE_FORMAT,
            level=logging.INFO,
        )
        if self.command == "setup":
            return True

        self.syncCache = synccache.SyncCache(self.config)
        if not self.syncCache.isValid() and self.command in (
            "show",
            "search",
            "install",
            "reinstall",
        ):
            print(" Please run './gsm update' to create an initial package index")
            return False

        self.localCache = localcache.LocalCache(
            self.config.localDir, self.config, logging
        )
        self.localCache.setSyncCache(self.syncCache)

        if self.command not in ("migrate", "switch-os"):
            if not utils.checkOS(self.config):
                return False
            return self.config.checkMigration()

        return True

    def setMaxlenPackageFullName(self, packages, length=MAX_LEN_PACKAGE_FULL_NAME):
        if self.syncCache.isValid():

            utils.setMaxlenPackageFullName(
                packages, self.localCache, self.syncCache, self.config
            )
        else:
            self.config.maxlenPackageFullName = length

    def filterPackages(self, packages):
        if not self.excludePackages:
            return packages

        names = self.excludePackages.split(",")
        for name in names:
            if name in packages:
                packages.remove(name)

        return packages

    def removeObsoletePackages(self, packages):
        obsoletePackages = []
        l = packages[:]
        for name in packages:
            if not self.syncCache.contains([name], False):
                obsoletePackages.append(name)
                l.remove(name)
                if not self.localCache.getPackage(name):
                    obsoletePackages.remove(name)
                    l.append(name)

        if obsoletePackages:
            print(
                "The following package(s) are no longer available:\n"
                f" {', '.join(obsoletePackages)}"
            )

            for name in obsoletePackages:
                response = utils.getDataFromUrl(
                    urljoin(GSM_INFO_OBSOLETE_PACKAGES_URL, name), self.config, True
                )
                if response:
                    print(response)
            if checkInput("\nDo you want to continue?", "n", "yn"):
                return []

        return l

    def removeHoldPackages(self, packages):
        holdPackages = []
        l = packages[:]
        for name in packages:
            if self.localCache.isHold(name):
                holdPackages.append(name)
                l.remove(name)
        if holdPackages:
            log.logInfo(
                f"The following package(s) set on hold : {', '.join(holdPackages)}"
            )

        return l

    def localInstall(self, packagesFiles):
        packages = []
        localPackages = {}

        for localpkg in packagesFiles:
            descFile = localpkg + DSC_EXTENSION

            if os.path.isfile(descFile) and os.path.isfile(localpkg):
                pkg = Package.fromFile(descFile, ".")
                pkg.dscFile = os.path.abspath(descFile)
                if (
                    pkg.os == self.config.os
                    and pkg.osVersion == self.config.os_version
                    and pkg.release == self.config.release
                    and pkg.arch == self.config.arch
                    or pkg.type in (PT.DATA, PT.CONFIG)
                ):
                    localPackages[pkg.name] = [pkg]
                    packages.append(pkg.name)
                elif (
                    pkg.arch == NO_ARCH
                    and pkg.os == ANY
                    and (pkg.osVersion in ("", ANY))
                ):
                    localPackages[pkg.name] = [pkg]
                    packages.append(pkg.name)
                else:
                    print(
                        f"SKIP : The {localpkg} has been built"
                        " for a different target system."
                    )
            else:
                if not os.path.isfile(descFile):
                    print(f"INFO : The file {descFile} does not exist.")

                if not os.path.isfile(localpkg):
                    print(f"INFO : The file {localpkg} does not exist.")

        try:
            installed_version = self.localCache.getPackage(pkg.name).version
        except Exception:
            installed_version = None

        if installed_version:
            if pkg.version < installed_version:
                print(
                    f"The package {pkg.name} is already installed in a "
                    f"newer version {installed_version}."
                    " You can try to remove the installed version "
                    "and restart the installation again."
                )
                return True

        self.syncCache.setIndex(localPackages)
        for localpkg in packagesFiles:
            self.config.downloadDir = os.path.dirname(os.path.abspath(localpkg))
            index = packagesFiles.index(localpkg)
            self.install(packages[index : index + 1], False, True)
            os.chdir(self.config.baseDir)

        return True

    def selfUpdate(self):
        value = self.config.assumeYes
        self.config.assumeYes = True
        self.install((["gsm"]), False)
        self.config.assumeYes = value

    def installLicenses(self, packages):
        self.setMaxlenPackageFullName(packages)

        ret = True
        for name in packages:
            pkg = self.localCache.getPackage(name)
            if not pkg and self.config.forceDownload:
                pkgs = self.syncCache.getPackages(name)
                if not pkgs:
                    log.logError(
                        f"Unable to locate package {name} for"
                        f" {self.config.os}/{self.config.os_version} (Release:"
                        f" {self.config.release})"
                    )
                    continue

                pkg = self.syncCache.getPackages(name)[0]

            if not pkg:
                print(f"Package not installed: {name}")
                continue

            # pkg = pkgs[0]
            if pkg.license == "None":
                continue

            if self.config.downloadOnly:
                if not package.downloadLicense(pkg, self.config):
                    ret = False
            else:

                if not package.downloadLicense(
                    pkg, self.config
                ) or not pkg.installLicense(self.config):
                    ret = False

        return ret

    def install(self, packages, force, localInstall=False):
        packages = self.filterPackages(packages)[:]
        if not packages:
            print("\nNo matching packages found.")
            return False

        packages.sort()

        packages = self.removeObsoletePackages(packages)[:]
        if not packages:
            print("\nAll selected packages became obsolete.")
            return False

        packages = self.removeHoldPackages(packages)[:]
        if not packages:
            print("\nAll selected packages are marked as hold.")
            return False

        if not self.syncCache.contains(packages):
            return False

        if not (len(packages) == 1 and packages[0] == "gsm"):
            ret, retCode = package.updateCheck(
                self.localCache, self.syncCache, self.config
            )

            if ret:
                self.selfUpdate()

            if retCode == 10:
                sys.exit(10)

        if not self.config.assumeYes:
            packages = self.localCache.checkOptionalInstall(packages)

        self.setMaxlenPackageFullName(packages)

        if not self.config.offline and not self.config.updateOrDowngrade:
            failed = utils.checkUrlAccessProgress(self.config, self.syncCache, packages)

            if failed:
                print(
                    "\nThe following packages are not activated in this repository:"
                    f" {', '.join(failed)}"
                )
                print("Please get in touch with info@gempa.de")
                q = "Do you want to (s)kip those packages, (c)ancel the installation entirely?"
                answer = utils.getInput(q, "s", "sc")
                if answer == "s":
                    for name in failed:
                        packages.remove(name)
                elif answer == "c":
                    sys.exit(0)
                print("")

                if not packages:
                    print("No package remain to be installed.")
                    return False

        failed = []
        for name in packages:
            pkg = self.syncCache.getPackages(name)[0]

            if pkg.license != "None":
                if not self.config.offline:
                    if not package.downloadLicense(pkg, self.config):
                        failed.append(name)

            if not self.localCache.isInstallable(pkg):
                failed.append(name)

        if failed:
            print(
                "\nThe licenses of the following packages are expired or are missing:"
                f" {', '.join(failed)}"
            )
            q = (
                "Do you want to (s)kip those packages, (i)gnore the missing licenses or"
                " (c)ancel the installation entirely?"
            )
            answer = utils.getInput(q, "s", "sic")
            if answer == "s":
                for name in failed:
                    packages.remove(name)
            elif answer == "c":
                sys.exit(0)
            print("")

            if not packages:
                print("No package with a valid licence file remain to be installed.")
                return False

        installed = []
        for name in packages:
            pkg = self.syncCache.getPackages(name)[0]
            if self.config.fromUpdate:
                show = False
            else:
                show = True
            if not force and self.localCache.isInstalled(pkg, show):
                if not self.config.fromUpdate:
                    installed.append(name)

        for name in installed:
            packages.remove(name)

        if not packages:
            print(
                "\nAll packages are already installed. Use 'reinstall' to force"
                " re-installation."
            )
            return False

        if installed and not self.config.fromUpdate:
            print(
                "\nThe following packages are installed in their latest version:"
                f" {', '.join(installed)}\nUse 'reinstall' to force re-installation.\n"
            )

        runCheckSum = []
        fileNotExist = []
        fileNotExistCrt = []
        fileNotExistLicenseTxt = []
        runCheckSumFailedDownloadAgain = []

        for name in packages:
            pkgSync = self.syncCache.getPackages(name)[0]
            pkg = self.localCache.getPackage(name)

            if not pkg or (pkg.version < pkgSync.version):
                pkg = pkgSync

            if (pkg and pkg.checksum == pkgSync.checksum) or (
                pkg and not pkg.checksum == pkgSync.checksum
            ):
                fileName = os.path.join(self.config.downloadDir, pkg.fileName)
                if not self.config.updateOrDowngrade and not force:
                    if os.path.exists(fileName):
                        if not utils.hashfile(
                            self.config, fileName, pkgSync.checksum, pkg
                        ):
                            if os.path.isfile(fileName):
                                runCheckSumFailedDownloadAgain.append(pkg.name)
                                os.remove(fileName)
                        else:
                            runCheckSum.append(pkg.name)
                    else:
                        fileNotExist.append(pkg)
                else:
                    runCheckSum.append(pkg.name)

                if pkg.license == "Gempa":
                    if pkg.licenseFile:
                        licenseFile = pkg.licenseFile
                    else:
                        licenseFile = pkg.name

                    fileNameCrt = os.path.join(
                        self.config.downloadLicenseDir, licenseFile + CRT_EXTENSION
                    )
                    if not os.path.exists(fileNameCrt):
                        fileNotExistCrt.append(fileNameCrt)

                    fileNameLicenseTxt = os.path.join(
                        self.config.downloadLicenseDir,
                        licenseFile + LICENSE_TXT_EXTENSION,
                    )
                    if not os.path.exists(fileNameLicenseTxt):
                        fileNotExistLicenseTxt.append(fileNameLicenseTxt)
        if not self.config.offline or not localInstall:
            if runCheckSumFailedDownloadAgain:
                print(
                    f"\nThe checksum check for the following packages has failed:"
                    f" {', '.join(runCheckSumFailedDownloadAgain)}."
                    " The download is forced for these packages.\n"
                )
        if self.config.downloadOnly or len(self.args) == 0:
            if (not self.installRequested and not self.config.assumeYes) or (
                not self.installRequested and not self.config.assumeYes
            ):
                print(
                    f"The following packages are selected to install: {', '.join(packages)}\n"
                )

        if self.config.offline or localInstall:
            downloaded = packages
            failed = set(downloaded) - set(runCheckSum)

            if fileNotExist:
                for pkg in fileNotExist:
                    print(
                        f"The {pkg.fileName} file does not exist in the"
                        f" {self.config.downloadDir} directory."
                    )
                    if pkg.name in failed:
                        failed.remove(pkg.name)

            if failed:
                print(
                    "\nChecksum validation for the following packages failed:"
                    f" {', '.join(failed)}"
                )

            if fileNotExistCrt:
                for name in fileNotExistCrt:
                    print(
                        f"The {name} file does not exist in the"
                        f" {self.config.downloadLicenseDir} directory."
                    )
            if fileNotExistLicenseTxt:
                for name in fileNotExistLicenseTxt:
                    print(
                        f"The {name} file does not exist in the"
                        f" {self.config.downloadLicenseDir} directory."
                    )

            if fileNotExist or failed or fileNotExistCrt or fileNotExistLicenseTxt:
                return True

        else:
            downloaded = self.syncCache.downloadPackages(packages)
            failed = set(packages) - set(downloaded)
            if failed:
                for name in failed:
                    packages.remove(name)

                if not packages:
                    print("\nThe download of all packages failed.")
                    return False

                print(
                    f"\nThe download of the following packages failed: {', '.join(failed)}"
                )
                if not checkInput("Do you want to continue?", "y", "yn"):
                    return False

            failed = []

        downloaded = set(downloaded) - set(runCheckSum)
        for name in downloaded:
            if not self.syncCache.checksumCheck(name):
                failed.append(name)

        if failed:
            print(
                "\nChecksum validation for the following packages failed:"
                f" {', '.join(failed)}"
            )
            print("\nTry to run ./gsm update to refresh the package index.")

            if self.config.downloadOnly:
                return False

            q = (
                "\nDo you want to (s)kip those packages, (i)gnore the checksum errors or"
                " (c)ancel the installation entirely?"
            )
            answer = getInput(q, "s", "sic")
            if answer == "s":
                for name in failed:
                    packages.remove(name)
            elif answer == "c":
                sys.exit(0)

            if not packages:
                print("No package with a valid checksum remain to be installed.")
                return False

            print("")

        if self.config.downloadOnly:
            return True

        ret, showChangeLog = package.showChangeLog(
            packages,
            self.config,
            self.localCache,
            self.syncCache,
            self.config.offline,
            True,
        )

        if ret and showChangeLog and not self.config.assumeYes:
            if not checkInput(
                "There are important changes needed after the installation!"
                " Do you want to continue?",
                "Y",
                "yn",
            ):
                return False

        if not showChangeLog:
            if not self.config.assumeYes:
                if not self.installRequested:
                    print(f"{log.getLogPrefix('INSTALL')}{', '.join(packages)}\n")
                    # if not self.installRequested:
                    if not checkInput("Do you want to continue?", "Y", "yn"):
                        return False

        installed = []
        success = True

        for name in packages:
            ret, summary = self.localCache.installPackage(name, localInstall)
            if ret == 0:
                installed.append(summary)
            else:
                success = False

        # Sets the Python interpreter for the selected SeisComP release
        if self.localCache.getPackage(SC) or self.localCache.getPackage(SC3):
            utils.createWrapper(self.config)

        print(f"\n{len(installed)}/{len(packages)} packages have been installed:\n")

        for summary in installed:
            pkg = self.localCache.getPackage(summary.name)
            print(
                f"{pkg.name} ({pkg.version}) to {summary.path}:"
                f" {summary.fileCount} file(s), {utils.humanbytes(pkg.installSize)} "
            )

        if os.path.isdir(self.config.libDir):
            fileList = makelinks.getFilesList(self.config.libDir)
            linkList = makelinks.getLinkList(fileList)
            ret = makelinks.createSymLinks(linkList, self.config.libDir)
            return success and ret

        return success

    def remove(self, packages):
        packages = self.filterPackages(packages)
        packages.sort()
        failed = []
        for name in packages:
            pkg = self.localCache.getPackage(name)
            if not pkg:
                log.logInfo(f"{name} is not installed ... ")
                failed.append(name)

        for name in failed:
            packages.remove(name)

        if not packages:
            print("")
            print("Nothing todo ...")
            return True

        if not self.config.assumeYes:
            packages = self.localCache.checkOptionalRemove(packages)

        if not self.config.assumeYes:
            print(f"{log.getLogPrefix('REMOVE')}{', '.join(packages)}\n")
            if checkInput("Do you want to continue?", "n", "yn"):
                return False

        ret = True
        removedPackages = []

        self.setMaxlenPackageFullName(packages)

        members = utils.sorting(packages)
        count = 0
        progress = tqdm(
            members,
            ascii=True,
            bar_format=utils.getTerminalSizeBarFormat(self.config),
        )

        for member in progress:
            pkg = self.localCache.getPackage(member)
            pkgName, _ = utils.getPackageName(self.localCache.getPackage(member))
            progress.set_description(
                f"{log.getLogPrefix('REMOVE')}{pkgName: <{self.config.maxlenPackageFullName:}}"
            )
            count = count + 1
            done = f"DONE ({len(progress)}/{count})"
            _, pkg.fileCount = self.localCache.getRemovableFiles(pkg)

            if localcache.PKG_SUCCESS == self.localCache.removePackage(member, False):
                removedPackages.append(pkg)
            else:
                ret = False
            if count == len(members):
                progress.set_description(
                    f"{log.getLogPrefix(REMOVE_OK)}"
                    f"{done: <{self.config.maxlenPackageFullName:}}"
                )
        print(f"\n{len(removedPackages)}/{len(packages)} packages have been removed:\n")

        for summary in removedPackages:
            print(
                f"{summary.name} ({summary.version}) from {summary.path}:"
                f" {summary.fileCount} file(s),"
                f" {utils.humanbytes(summary.installSize)} "
            )

        fileList = makelinks.getFilesList(self.config.libDir)
        linkList = makelinks.getLinkList(fileList)

        return ret and makelinks.createSymLinks(linkList, self.config.libDir)

    def parseCommandLine(self, initStage=False):
        try:
            opts, args = getopt.gnu_getopt(
                self.argv[1:],
                "hu:p:r:dyof",
                [
                    "help",
                    "user=",
                    "assumeyes",
                    "password=",
                    "release=",
                    "arch=",
                    "url=",
                    "noverify",
                    "downloadonly",
                    "offline",
                    "downloaddir=",
                    "debug",
                    "installdir=",
                    "showurls",
                    "exclude=",
                    "force",
                ],
            )
        except getopt.GetoptError as err:
            # will print something like "option -a not recognized"
            print(str(err))
            return False

        self.command = None
        for o, a in opts:
            if o in ["-h", "--help"]:
                self.command = "help"
                return True

            if initStage:
                continue

            if o in ["-u", "--user"]:
                self.config.user = a
            elif o in ["-y", "--assumeyes"]:
                self.config.assumeYes = True
            elif o in ["-p", "--password"]:
                self.config.password = a
            elif o in ["-r", "--release"]:
                self.config.release = a
            elif o in ["--arch"]:
                self.config.arch = a
            elif o in ["--url"]:
                self.config.baseURL = a
            elif o in ["-d", "--downloadonly"]:
                self.config.downloadOnly = True
            elif o in ["-o", "--offline"]:
                self.config.offline = True
            elif o in ["--downloaddir"]:
                self.config.downloadDir = a
            elif o in ["--installdir"]:
                self.config.installDir = a
            elif o in ["--debug"]:
                self.config.debug = True
            elif o in ["--showurls"]:
                self.config.showUrls = True
            elif o in ["--exclude"]:
                self.excludePackages = a
            elif o in ["-f", "--force"]:
                self.config.forceDownload = True
            else:
                print(f"unhandled option: {o}", file=sys.stderr)
                return False

        if not args:
            print("no arguments given", file=sys.stderr)
            return False

        self.command = args[0].lower()
        self.args = args[1:]

        return True

    def run(self):
        return self.init() and self.handleCommand(self.command)


def main():
    app = App(len(sys.argv), sys.argv)
    return not app.run()


if __name__ == "__main__":
    try:
        sys.exit(main())
    except KeyboardInterrupt:
        print("Interrupted (Ctrl+C)")
        sys.exit(signal.SIGINT)


# vim: ts=4 noet
