############################################################################
# Copyright (C) 2019 by gempa GmbH                                         #
#                                                                          #
# All Rights Reserved.                                                     #
#                                                                          #
# NOTICE: All information contained herein is, and remains                 #
# the property of gempa GmbH and its suppliers, if any. The intellectual   #
# and technical concepts contained herein are proprietary to gempa GmbH    #
# and its suppliers.                                                       #
# Dissemination of this information or reproduction of this material       #
# is strictly forbidden unless prior written permission is obtained        #
# from gempa GmbH.                                                         #
#                                                                          #
#  Author: Enrico Ellguth, Ralph Henneberger                               #
#  Email: enrico.ellguth@gempa.de, ralix@gempa.de                          #
############################################################################

import glob
import logging
import os
import shutil
import sys
from time import gmtime, strftime
from datetime import datetime
from pathlib import Path


import pkg_resources

from libs import package, utils, log
from libs.cache import custominstall
from libs.const import (
    NIGHTLY,
    PT,
    SC3,
    GREENSFUNCTIONS,
    FILE_ENCODING,
    FORMAT,
    DEFAULT_TIME_FORMAT,
    INSTALL_DESC,
    INSTALL_FILES,
    DEFAULT_DATE_FORMAT,
    MAX_LEN_PACKAGE_FULL_NAME,
    HOLD_FILE,
    MBTILES_EXTENSION,
    GRD_EXTENSION,
    CRT_EXTENSION,
    LICENSE_TXT_EXTENSION,
    ARCHIVE,
)
from libs.package import Package

PKG_SUCCESS = 0
PKG_FAILED = 1
PKG_NOT_FOUND = 2
PKG_INVALID_VERSION = 3
PKG_IS_ALREADY_INSTALLED = 4


class Summary(object):
    def __init__(self, name, fileCount, path):
        self.name = name
        self.fileCount = fileCount
        self.path = path


class LocalCache(object):
    def __init__(self, path, config, logger):
        self.index = {}
        self.path = path
        self.index = self.read()
        self.config = config
        self.logging = logger
        self.syncCache = None
        self.holdFile = os.path.join(self.config.localDir, HOLD_FILE)

        self.holdList = []
        if os.path.isfile(self.holdFile):
            try:
                with open(self.holdFile, "r", encoding=FILE_ENCODING) as f:
                    self.holdList = [l.strip() for l in f.readlines()]
            except IOError as e:
                log.logError(f"could not read package hold list: {e}")

    def setSyncCache(self, cache):
        self.syncCache = cache

    def read(self):
        if not os.path.isdir(self.path):
            return self.index

        modules = [
            name
            for name in os.listdir(self.path)
            if os.path.isdir(os.path.join(self.path, name))
        ]
        for fileName in modules:
            descFile = os.path.join(self.path, fileName, INSTALL_DESC)
            try:
                pkg = Package.fromFile(descFile, self.path)
                self.index[pkg.name] = [pkg]
            except BaseException as e:
                log.logError(f"Could not read package description '{descFile}': {e}")

        return self.index

    def getFileInfo(self):
        fileInfo = {}
        allFiles = glob.glob(os.path.join(self.config.localDir, "*"))
        for fileName in allFiles:
            with open(
                os.path.join(fileName, INSTALL_FILES), "r", encoding=FILE_ENCODING
            ) as f:
                for line in f:
                    line = line.rstrip()
                    if line in fileInfo:
                        fileInfo[line] += 1
                    else:
                        fileInfo[line] = 1

        return fileInfo

    def getRemovableFiles(self, pkg):
        path = os.path.join(self.config.localDir, pkg.name)
        fileInfo = {}
        fileList = []
        fileCount = 0

        fileInfo = self.getFileInfo()

        if pkg.type == PT.DATA and not pkg.filePath == ARCHIVE:
            removeDir = self.config.dataDir
            pkg.path = self.getInstallPath(pkg)
        elif pkg.filePath == ARCHIVE:
            removeDir = self.config.dataDir
            pkg.path = removeDir
        elif pkg.type == PT.CONFIG:
            removeDir = os.path.join(self.config.installDir, self.config.rootDir)
            pkg.path = removeDir
        elif pkg.type == PT.GSM:
            removeDir = self.getInstallPath(pkg)
        else:
            removeDir = os.path.join(self.config.installDir, self.config.rootDir)
            pkg.path = os.path.join(self.getInstallPath(pkg), self.config.rootDir)

        # add all used directories
        with open(os.path.join(path, INSTALL_FILES), "r", encoding=FILE_ENCODING) as f:
            for line in f:
                line = line.rstrip()
                if fileInfo.get(line) != 1:
                    continue

                fileName = os.path.join(removeDir, line)
                if not os.path.isfile(fileName) and not os.path.isdir(fileName):
                    log.logError(
                        f"Could not remove {fileName}: No such file or directory."
                    )
                fileList.append(fileName)
                if os.path.isfile(os.path.join(removeDir, line)):
                    fileCount += 1

        if pkg.license != "None":
            try:
                crtFile = os.path.join(self.config.licenseDir, pkg.name + CRT_EXTENSION)
                if os.path.isfile(crtFile):
                    fileList.append(crtFile)
                    fileCount += 1

                txtFile = os.path.join(
                    self.config.licenseDir, pkg.name + LICENSE_TXT_EXTENSION
                )
                if os.path.isfile(txtFile):
                    fileList.append(txtFile)
                    fileCount += 1

            except Exception:
                pass
        if pkg.type == PT.DATA:
            if pkg.fileName.endswith(GRD_EXTENSION) or pkg.fileName.endswith(
                MBTILES_EXTENSION
            ):
                fileCount = 1
            else:
                fileCount = 0
                for _, _, files in os.walk(pkg.path):
                    fileCount += len(files)

        return fileList, fileCount

    def getIndex(self):
        return self.index

    def isInstalledToday(self, installdate):
        try:
            installDate = datetime.strptime(installdate, DEFAULT_DATE_FORMAT).strftime(
                DEFAULT_TIME_FORMAT
            )
        except ValueError:
            return False

        today = datetime.now().strftime(DEFAULT_TIME_FORMAT)
        return installDate == today

    def isInstalled(self, pkg, show=False):
        try:
            item = self.index.get(pkg.name, [])[0]

            # Do not update gsm dev version
            if item.version == "VERSION":
                return True

            if item.version == NIGHTLY:
                version = item.version
            else:
                version = pkg_resources.parse_version(item.version)

            if pkg.version == NIGHTLY:
                pkgVersion = pkg.version
            else:
                pkgVersion = pkg_resources.parse_version(pkg.version)

            checksum = item.checksum
            installDate = item.installDate
            if version > pkgVersion:
                return True

            if version == pkgVersion:
                if self.config.release == NIGHTLY:
                    if not pkg.type == PT.APP:
                        if version == pkgVersion:
                            return True
                    return False

                if show:
                    if checksum != pkg.checksum and self.isInstalledToday(installDate):
                        log.logInfo(
                            "Force reinstall of"
                            f" {pkg.name} {self.index.get(pkg.name, [])[0].version} (invalid"
                            " checksum)"
                        )
                        return False

                    log.logSkip(
                        f"Latest version of {pkg.name} is already installed:"
                        f" {self.index.get(pkg.name, [])[0].version} "
                    )
                return True

            if pkg.type == PT.CONFIG:
                if show:
                    log.logSkip(
                        f"The package {pkg.name} can't be installed because it might"
                        " overwrite user-defined configuration changes."
                    )
                return True

        except IndexError:
            pass

        return False

    def writeHoldList(self):
        try:
            with open(self.holdFile, "w", encoding=FILE_ENCODING) as f:
                f.write("\n".join(self.holdList))
        except IOError as e:
            log.logError(f"Could not write package hold list: {e}")
            return False
        if os.stat(self.holdFile).st_size == 0:
            os.remove(self.holdFile)
        return True

    def addHoldPackage(self, name):
        if name in self.holdList:
            log.logError(f"The package {name} is already on hold")
            return False

        self.holdList.append(name)
        if not self.writeHoldList():
            return False

        log.logInfo(f"The package {name} was marked as hold")
        return True

    def removeHoldPackage(self, name):
        if name not in self.holdList:
            log.logError(f"The package {name} is not on hold")
            return False

        self.holdList.remove(name)
        if not self.writeHoldList():
            return False

        log.logInfo(f"The package {name} was removed from hold")
        return True

    def showHoldPackages(self):
        if not os.path.exists(self.holdFile):
            log.logInfo("No packages on hold")
            return True

        print("Packages on hold:")
        for name in self.holdList:
            pkg = self.getPackage(name)
            if not pkg:
                continue

            if not self.syncCache.contains([name], False):
                errorText = " (No longer available)"
            else:
                errorText = ""
            print(f"  {pkg.name} ({pkg.version}){errorText}")

        return True

    def isHold(self, name):
        return name in self.holdList

    def getPackage(self, fullname):
        try:
            name = fullname.split("=")[0]
            pkg = self.index.get(name, [])[0]

            if "=" not in fullname:
                return pkg
            if fullname != pkg.fullName():
                return None
            return pkg

        except IndexError:
            return None

    def getLicenseInfo(self, config, pkg):
        license_name = package.getLicenseFileName(pkg)
        cert = utils.readCertificate(license_name, config)
        if cert:
            _, _, updatesUntil, notValidAfter = utils.getLicenseTimeDelta(cert)
            return updatesUntil, notValidAfter

        return "None", "None"

    def getMaxLen(self, headerValue, value):
        maxlenValue = len(max(value, key=len))
        maxlenValue = min(max(maxlenValue, len(headerValue)), MAX_LEN_PACKAGE_FULL_NAME)

        return headerValue, maxlenValue

    def infoPackagesShort(self):
        installed_packages = [pkg for pkg in self.index.keys() if pkg != "gsm"]
        installed_packages.sort()
        print(f"Installed packages: {' '.join(installed_packages)}")

        return True

    def infoPackages(self, packages=None):
        if not packages:
            headerPackages, maxlenPackage = self.getMaxLen(
                "Installed packages", self.index.keys()
            )

            versions = []
            datesValid = []
            datesUpdate = []

            for name in self.index.keys():
                pkg = self.getPackage(name)
                versions.append(pkg.version)
                dateUpdate, dateValid = self.getLicenseInfo(self.config, pkg)
                datesValid.append(dateValid)
                datesUpdate.append(dateUpdate)

            headerVersion, maxlenVersion = self.getMaxLen("Version", versions)
            headerDates, maxlenDates = self.getMaxLen("Valid until", datesValid)
            _, maxlenUntil = self.getMaxLen("Updates until", datesUpdate)

            print(
                f"{headerPackages: <{maxlenPackage:}}  {headerVersion: <{maxlenVersion:}}  "
                f"{headerDates: <{maxlenDates:}}  Updates until"
            )
            print(
                f"{''.ljust(maxlenPackage, '-')}  {''.ljust(maxlenVersion, '-')}  "
                f"{''.ljust(maxlenDates, '-')}  -------------"
            )
            for name in sorted(self.index):
                pkg = self.getPackage(name)
                updatesUntil, notValidAfter = self.getLicenseInfo(self.config, pkg)

                if len(self.syncCache.getIndex()) == 0:
                    errorText = ""
                else:
                    errorText = (
                        ""
                        if self.syncCache.contains([name], False)
                        else " (no longer available)"
                    )

                print(
                    f"{pkg.name: <{maxlenPackage:}}  {pkg.version: <{maxlenVersion:}}"
                    f"  {notValidAfter: <{maxlenDates:}}  {updatesUntil: <{maxlenUntil}}"
                    f"  {errorText:}"
                )
            return True

        first = True
        for name in packages:
            # no new line for first package
            first = not first and print("")
            pkg = self.getPackage(name)
            if not pkg:
                print(f"Package not installed: {name}")
                continue

            print(name)
            print("".ljust(max(len(name), 10), "-"))
            pkg.showInfo(True)
            if not self.syncCache.contains([name], False):
                print("Info:          : No longer available")
        return True

    def createDesc(self, pkg, localInstall=False):
        if localInstall:
            dscFile = pkg.dscFile
        else:
            baseDirPath = self.syncCache.getPackagePath(pkg)
            if pkg.filePath or pkg.type == PT.CONFIG:
                dscFile = os.path.join(self.config.syncDir, baseDirPath, pkg.dscFile)

            else:
                dscFile = os.path.join(
                    self.config.syncDir, pkg.name, baseDirPath, pkg.dscFile
                )

        installDir = os.path.join(self.config.localDir, pkg.name)
        installDesc = os.path.join(installDir, INSTALL_DESC)

        if not utils.createDirectory(installDir):
            return False

        try:
            shutil.copyfile(dscFile, installDesc)
            with open(installDesc, "a+", encoding=FILE_ENCODING) as f:
                f.write(
                    f"\ninstalldate=('{strftime(DEFAULT_DATE_FORMAT, gmtime())}')\n"
                )
        except IOError as err:
            print(f"Failed to create package description: {err}")
            return False

        return True

    def createFiles(self, pkg, fileList):
        installDir = os.path.join(self.config.localDir, pkg.name)
        installFiles = os.path.join(installDir, INSTALL_FILES)
        if not utils.createDirectory(installDir):
            return False

        try:
            with open(installFiles, "w", encoding=FILE_ENCODING) as f:
                if pkg.type == PT.DATA and not pkg.filePath == ARCHIVE:
                    if GREENSFUNCTIONS in pkg.filePath:
                        f.write(pkg.filePath + "\n")
                        f.write(pkg.filePath + ".desc\n")
                    elif pkg.installPath:
                        f.write(pkg.installPath)
                    else:
                        f.write(pkg.filePath)

                    return True

                for member in fileList:
                    if pkg.type == PT.GSM:
                        f.write(f"{member.replace(SC3 + os.sep, '')}\n")
                    elif pkg.type == PT.CONFIG:
                        f.write(f"{member}\n")
                    elif pkg.filePath == ARCHIVE:
                        f.write(f"{os.path.join(ARCHIVE, Path(member))}\n")
                    else:
                        prefix = self.config.installPrefix()
                        if not member == prefix:
                            f.write(f"{member.replace(prefix + os.sep, '', 1)}\n")

        except OSError as err:
            print(f"Failed to create package file list: {err}")
            return False

        return True

    def getInstallPath(self, pkg):
        if pkg.type == PT.GSM:
            return os.path.split(os.path.abspath(os.getcwd()))[0]
        if pkg.type == PT.DATA:
            if pkg.installPath:
                return os.path.join(self.config.dataDir, pkg.installPath)

            return os.path.join(self.config.dataDir, pkg.filePath)

        if pkg.type == PT.CONFIG:
            return self.config.installPath

        return self.config.installDir

    def getSummaryPath(self, pkg):
        if pkg.type == PT.DATA:
            if pkg.installPath and not GREENSFUNCTIONS in pkg.filePath:
                return os.path.join(self.config.dataDir, pkg.installPath)

            return os.path.join(self.config.dataDir, pkg.filePath)

        installPath = self.getInstallPath(pkg)

        if pkg.type == PT.GSM:
            return os.path.join(installPath)

        if pkg.type == PT.CONFIG:
            return self.config.installPath

        # type = PT.APP
        return os.path.join(installPath, self.config.rootDir)

    def installPackage(self, name, localInstall=False):
        try:
            pkg = self.syncCache.getPackages(name)[0]
        except IndexError:
            log.logError(
                f"Unable to locate package {name} for"
                f" {self.config.os}/{self.config.os_version} (Release:"
                f" {self.config.release})"
            )
            return PKG_NOT_FOUND, None

        if pkg.format != package.FORMAT:
            if utils.splitVersionString(
                self.getPackage("gsm").version
            ) < utils.splitVersionString(FORMAT[pkg.format]):
                log.logError(
                    f"{pkg.dscFile} is not compatible ({pkg.format} != {package.FORMAT})\n"
                    f"Please use gsm version >= {FORMAT[pkg.format]}"
                )
                return PKG_INVALID_VERSION, None

        installPath = self.getInstallPath(pkg)

        if pkg.type not in (PT.CONFIG, PT.GSM):
            self.removePackage(pkg.name, False)

        fileName = os.path.join(self.config.downloadDir, pkg.fileName)
        currentDir = os.getcwd()

        if not utils.createDirectory(installPath):
            return PKG_FAILED, None

        if pkg.type == PT.CONFIG:
            existFileName = utils.checkFilesExist(fileName, self.config.installPath)
            if existFileName:
                log.logError(
                    f"{os.path.basename(existFileName)} from package {pkg.name} already exist"
                    f" in {os.path.join(installPath, os.path.dirname(existFileName))}"
                )
                return PKG_FAILED, None

        if pkg.installScript:
            ret, existFileName = utils.runCustomInstallScript(self, fileName, pkg)
            if not ret:
                log.logError(f"{pkg.installScript}' not found in {fileName}")
                return PKG_FAILED, None

            if existFileName:
                log.logError(
                    f"{os.path.basename(existFileName)} from package {pkg.name} exist"
                    f" in {os.path.join(installPath, os.path.dirname(existFileName))}"
                )
                return PKG_FAILED, None

        if pkg.filePath:
            if pkg.name == "gsm":
                if os.system(f"tar xf {fileName} --strip-components=1") == 0:
                    print("OK")
                else:
                    print("FAILED")
                    return PKG_FAILED, None

            extension = fileName[fileName.rfind(".") :]
            if (GRD_EXTENSION in extension) or (MBTILES_EXTENSION in extension):
                installDir = installPath
                installPath = os.path.join(installPath, pkg.fileName)
                if MBTILES_EXTENSION in extension:
                    if not utils.createDirectory(installDir):
                        return PKG_FAILED, None

                    workingDir = os.getcwd()
                    os.chdir(installDir)

                    try:
                        os.symlink(pkg.fileName, pkg.name + MBTILES_EXTENSION)
                    except FileExistsError as err:
                        log.logError(f"Package link exists: {err}")
                        return PKG_FAILED, None

                    os.chdir(workingDir)

        if pkg.name == "gsm":
            fileCount = 14
        else:
            ret, fileCount, fileList = utils.install(
                self.config, fileName, installPath, pkg
            )
            if fileCount == 0 or not ret:
                return PKG_FAILED, None

            self.createFiles(pkg, fileList)

        self.createDesc(pkg, localInstall)
        logging.info("%s was installed in version %s", pkg.name, pkg.version)
        custominstall.install(self, installPath, pkg)

        os.chdir(currentDir)
        if not pkg.license == "None":
            pkg.installLicense(self.config)
            fileCount = fileCount + 2

        self.index[pkg.name] = [pkg]

        summary = Summary(pkg.name, fileCount, self.getSummaryPath(pkg))

        return PKG_SUCCESS, summary

    def removePackage(self, name, info):
        pkg = self.getPackage(name)
        if not pkg:
            if info:
                log.logInfo(f"{name} is not installed ... ")
            return PKG_NOT_FOUND

        if info:
            print(f"{pkg.name} ({pkg.version}) ", end="")

        path = os.path.join(self.config.localDir, pkg.name)
        if not os.path.isdir(path):
            if info:
                print(f"{pkg.name} is not installed ... ")
            return PKG_NOT_FOUND

        try:
            dirInfo = {}
            removables, fileCount = self.getRemovableFiles(pkg)

            # add all used directories
            for line in removables:
                try:
                    if pkg.type == PT.DATA and not pkg.filePath == ARCHIVE:
                        if os.path.isfile(line):
                            os.remove(line)
                        else:
                            shutil.rmtree(line)
                        dirInfo[
                            os.path.dirname(
                                line.replace(self.config.dataDir + os.sep, "")
                            )
                        ] = 0

                        custominstall.remove(self, pkg)

                    else:
                        try:
                            os.remove(line)
                        except PermissionError:
                            log.logError(
                                f"Could not remove {line}: Permission denied \n"
                            )
                        dirInfo[
                            os.path.dirname(
                                line.replace(self.config.installDir + os.sep, "")
                            )
                        ] = 0

                except OSError:
                    pass

            if pkg.type not in (PT.CONFIG, PT.DATA):
                try:
                    if not os.listdir(self.config.licenseDir):
                        shutil.rmtree(self.config.licenseDir)
                except Exception:
                    pass

            shutil.rmtree(path)

            # remove all empty directories
            for key, _ in dirInfo.items():
                try:
                    path = key
                    while path:
                        if pkg.type == PT.DATA:
                            os.rmdir(os.path.join(self.config.dataDir, path))
                        else:
                            os.rmdir(os.path.join(self.config.installDir, path))
                        path = os.path.dirname(path)

                except OSError:
                    pass
        except Exception as ex:
            print(f"Could not remove package:  {str(ex)}")
            path = os.path.join(self.config.localDir, pkg.name)
            shutil.rmtree(path)
            info = False

        if info:
            if pkg.type == PT.DATA:
                print(
                    f"From {self.getSummaryPath(pkg)}:"
                    f" {utils.humanbytes(pkg.installSize)} "
                )
            else:
                print(
                    f"From {self.getSummaryPath(pkg)}: {fileCount} file(s),"
                    f" {utils.humanbytes(pkg.installSize)} "
                )
            logging.info("%s was deinstalled in version %s", pkg.name, pkg.version)
            del self.index[pkg.name]

            if not os.listdir(self.config.installDir):
                shutil.rmtree(self.config.installDir)

        return PKG_SUCCESS

    def isInstallable(self, pkg):
        if pkg.conflicts:
            for name in pkg.conflicts.split(" "):
                other = self.getPackage(name)
                if other:
                    log.logError(f"{pkg.name} is in conflict with {other.name} ... ")
                    sys.exit(1)
        if pkg.license == "None":
            return True

        version = 0
        if pkg.name == self.config.installPrefix():
            return True
        cert = utils.readCertificate(pkg.name, self.config)
        if cert:
            if cert.updatesUntil and cert.updatesUntil < pkg.version:
                log.logError(
                    f"The {pkg.name} license no longer allows updates. Updates until:"
                    f" {cert.updatesUntil} "
                )
                return False

            if not cert.updatesUntil:
                try:
                    version = int(self.config.release)

                except ValueError:
                    pass

                if version >= 4:
                    print(
                        f"No update until date found in {pkg.name} license. The update"
                        " information is required by gempa modules build for SeisComP"
                        " 4 and newer. Please contact support@gempa.de."
                    )
                    return False
        return True

    def checkOptionalInstall(self, packages, ask=None):
        optional = []

        for name in packages:
            pkg = self.syncCache.getPackages(name)[0]

            if pkg.optional:
                optionalPkgs = pkg.optional.split(" ")
                optionalPkgs = [
                    pkgName for pkgName in optionalPkgs if pkgName not in packages
                ]

                validOptionalPkgs = []
                for pkgName in optionalPkgs:
                    try:
                        pkgOptional = self.syncCache.getPackages(pkgName)[0]
                        if self.isInstallable(pkgOptional) and not self.isInstalled(
                            pkgOptional, False
                        ):
                            validOptionalPkgs.append(pkgName)
                    except IndexError:
                        pass
                for validOptionalPkg in validOptionalPkgs:
                    if validOptionalPkg not in optional:
                        optional.append(validOptionalPkg)

        if optional:
            if not ask:
                print(
                    "\nThe following packages are optional:",
                    f"{', '.join(optional)}\n",
                )

                if utils.checkInput("Do you want to add these packages?", "Y", "yn"):
                    packages.extend(optional)
            else:
                packages.extend(optional)

        return packages

    def checkOptionalRemove(self, packages):
        allOptional = []
        for name in sorted(self.getIndex()):
            pkg = self.getPackage(name)
            if pkg.optional:
                optionalPkgs = pkg.optional.split(" ")
                for name in optionalPkgs:
                    allOptional.append(name)

        optional = []

        for name in packages:
            pkg = self.getPackage(name)

            if pkg.optional:
                optionalPkgs = pkg.optional.split(" ")
                optionalPkgs = [
                    pkgName for pkgName in optionalPkgs if pkgName not in packages
                ]

                validOptionalPkgs = []
                for pkgName in optionalPkgs:
                    try:
                        pkgOptional = self.getPackage(name)
                        if self.isInstalled(pkgOptional, False):
                            validOptionalPkgs.append(pkgName)
                            if pkgName in allOptional:
                                allOptional.remove(pkgName)
                    except IndexError:
                        pass
                validOptionalPkgsTemp = validOptionalPkgs
                for validOptionalPkg in validOptionalPkgsTemp:
                    if (
                        validOptionalPkg not in optional
                        and validOptionalPkg not in allOptional
                    ):
                        optional.append(validOptionalPkg)
                    if validOptionalPkg in allOptional:
                        validOptionalPkgs.remove(validOptionalPkg)

                for packageName in packages:
                    if packageName not in validOptionalPkgs:
                        optional.append(packageName)
                        validOptionalPkgs.append(packageName)

                for x in packages:
                    pkgTmp = self.getPackage(x)
                    if pkgTmp.optional:
                        y = pkgTmp.optional.split(" ")
                        for _ in y:
                            if x in validOptionalPkgs:
                                validOptionalPkgs.remove(x)
                    else:
                        validOptionalPkgs.remove(x)

        if not optional:
            return packages

        if validOptionalPkgs:
            print(
                "\nThe following packages were optionally installed:",
                f"{', '.join(validOptionalPkgs)}\n",
            )

            if utils.checkInput("Do you want to remove these packages?", "Y", "yn"):
                packages.extend(validOptionalPkgs)

        return list(dict.fromkeys(packages))
