############################################################################
# Copyright (C) 2019 by gempa GmbH                                         #
#                                                                          #
# All Rights Reserved.                                                     #
#                                                                          #
# NOTICE: All information contained herein is, and remains                 #
# the property of gempa GmbH and its suppliers, if any. The intellectual   #
# and technical concepts contained herein are proprietary to gempa GmbH    #
# and its suppliers.                                                       #
# Dissemination of this information or reproduction of this material       #
# is strictly forbidden unless prior written permission is obtained        #
# from gempa GmbH.                                                         #
#                                                                          #
#  Author: Enrico Ellguth, Ralph Henneberger                               #
#  Email: enrico.ellguth@gempa.de, ralix@gempa.de                          #
############################################################################

import glob
import os
import re
import shutil
import textwrap

from natsort import natsorted

from libs import package, utils, log
from libs.const import LINE_WIDTH, PT, EXTEND_DIR, NO_ARCH, DSC_EXTENSION

from libs.package import Package


class SyncCache(object):
    def __init__(self, config):
        self.index = {}
        self.config = config
        self._isValid = self._read()

    def isValid(self):
        return self._isValid

    def _read(self):
        def getVersion(pkg):
            return pkg.version

        if not os.path.exists(self.config.syncDir):
            return False

        # create a list of all modules which are subdirectories of the sync
        # directory
        try:
            modules = sorted(
                filter(
                    lambda x: x != EXTEND_DIR
                    and os.path.isdir(os.path.join(self.config.syncDir, x)),
                    os.listdir(self.config.syncDir),
                )
            )
        except IOError as e:
            log.logError(f"Could not read sync directory: {e}")
            return False

        # regular modules
        for module in modules:
            packages = []
            moduleDir = os.path.join(self.config.syncDir, module)

            for relPath in (self.config.archDir, self.config.noArchDir):
                d = os.path.join(moduleDir, relPath)
                if not os.path.isdir(d):
                    continue

                for fileName in os.listdir(d):
                    if not fileName.endswith(DSC_EXTENSION):
                        continue

                    descFile = os.path.join(d, fileName)
                    packages.append(Package.fromFile(descFile, relPath))

            packages = natsorted(packages, key=getVersion, reverse=True)
            self.index[module] = packages

        # data and configuration packages
        if not os.path.isdir(self.config.extendDir):
            return True

        for relPath, _, files in os.walk(self.config.extendDir):
            packages = []
            module = os.path.basename(relPath)
            for fileName in files:
                if fileName.endswith(DSC_EXTENSION):
                    descFile = os.path.join(self.config.extendDir, relPath, fileName)
                    pkg = Package.fromFile(descFile, relPath)
                    if pkg.type == "app":
                        pkg.directory = ""
                    packages.append(pkg)

            packages = natsorted(packages, key=getVersion, reverse=True)
            # Ignore parent directories
            if packages:
                self.index[module] = packages

        return True

    def updateIndex(self):
        if not utils.createDirectory(self.config.syncDir):
            return False

        url = f"{self.config.dbUrl}/{self.config.dbFileName}"
        dbFile = os.path.join(self.config.syncDir, self.config.dbFileName)

        if utils.download(url, dbFile, self.config.dbFileName, self.config):
            try:
                allFiles = glob.glob(os.path.join(self.config.syncDir, "*"))
                excludedFiles = glob.glob(
                    os.path.join(self.config.syncDir, self.config.dbFileName)
                )
                for fileName in allFiles:
                    if fileName not in excludedFiles:
                        shutil.rmtree(fileName)

            except OSError as err:
                log.logFailed(f"{err}")
                return False

            if not utils.install(
                self.config, dbFile, self.config.syncDir, self.config.dbFileName
            ):
                return False

            if not self._read():
                return False

        return True

    def setIndex(self, localIndex):
        self.index = localIndex

    def getIndex(self):
        return self.index

    def contains(self, checkPackages, info=True):
        not_found = []
        for name in checkPackages:
            if not self.getPackages(name):
                not_found.append(name)

        if not_found:
            if info:
                log.logError(
                    f"Unable to locate package {', '.join(not_found)} for"
                    f" {self.config.os}/{self.config.os_version} (Release:"
                    f" {self.config.release})"
                )
            return False

        return True

    def getPackages(self, name):
        if name.find("=") >= 0:
            toks = name.split("=")
            pkgs = self.index.get(toks[0], [])
            for pkg in pkgs:
                if pkg.version == toks[1]:
                    return [pkg]
            return None

        return self.index.get(name, [])

    def infoPackages(self, packages=None, showAll=False):

        if not packages:
            header = "Available packages"

            # maximum package name length
            max_name_len = len(max(self.index.keys(), key=len))
            max_name_len = min(max(max_name_len, len(header)), 30)

            # minimum columns required for table view
            min_table_cols = max_name_len + 25

            # determine available TTY columns
            try:
                cols = int(os.popen("stty size", "r").read().split()[1])
            except Exception:
                cols = LINE_WIDTH

            # to few TTY colums for table view: print description in new line
            # indented by maximum 4 chars
            if cols < min_table_cols:
                print(header)
                print("".ljust(len(header), "-"))
                indent = "".ljust(min(cols, 4))
                for name in sorted(self.index):
                    if not self.getPackages(name):
                        continue
                    desc = (
                        self.index[name][0].desc.strip()
                        if self.index.get(name)
                        else None
                    )
                    print(name)
                    if desc:
                        print(
                            textwrap.fill(
                                desc,
                                width=cols,
                                initial_indent=indent,
                                subsequent_indent=indent,
                            )
                        )
            # enough space for table view
            else:
                print(f"{header: <{max_name_len:}}  Description")
                print(f"{''.ljust(max_name_len, '-')}  -----------")
                indent = "".ljust(max_name_len + 2)
                for name in sorted(self.index):
                    if not self.getPackages(name):
                        continue
                    desc = (
                        self.index[name][0].desc.strip()
                        if self.index.get(name)
                        else None
                    )
                    if desc:
                        line = f"{name: <{max_name_len:}}  {desc}"
                        try:
                            print(
                                textwrap.fill(
                                    line, width=cols, subsequent_indent=indent
                                )
                            )
                        except UnicodeEncodeError:
                            print(name)
                    else:
                        print(name)

            return True

        first = True
        for name in packages:
            if name not in self.index:
                # no new line for first package
                first = not first and print("")
                print(f"Package not found: {name}")
                continue

            pkgs = self.index.get(name, [])
            for i, pkg in enumerate(pkgs, start=1):
                first = not first and print("")
                header = name
                if showAll:
                    header += f" [{i}/{len(pkgs)}]"
                print(header)
                print("".ljust(max(len(header), 10), "-"))
                pkg.showInfo(False)
                if not showAll:
                    break

        return True

    def search(self, patterns):
        regexes = []
        for pattern in patterns:
            regexes.append(re.compile(pattern, re.IGNORECASE))

        matchedPkgs = {}
        for name in self.index:
            if all(regex.match(name) for regex in regexes):
                try:
                    pkg = self.index.get(name)[0]
                    matchedPkgs[name] = pkg
                except IndexError:
                    pass

        for name in sorted(self.index):
            try:
                pkgs = self.index.get(name, [])
                for pkg in pkgs:
                    for keyword in pkg.keywords.split():
                        if all(regex.search(keyword) for regex in regexes):
                            matchedPkgs[name] = pkg
            except IndexError:
                pass

        if not matchedPkgs:
            return False

        for name, pkg in matchedPkgs.items():
            print(f"{pkg.name} - {pkg.desc}")

        return True

    def checksumCheck(self, name):
        if not utils.createDirectory(self.config.downloadDir):
            return False

        try:
            pkg = self.getPackages(name)[0]
        except IndexError:
            log.logError(f"could not find package in index: {name}")
            return False

        fileName = os.path.join(self.config.downloadDir, pkg.fileName)

        if not utils.hashfile(self.config, fileName, pkg.checksum, pkg):
            return False

        return True

    def downloadPackages(self, packages):
        downloaded = []

        if not utils.createDirectory(self.config.downloadDir):
            return downloaded

        for name in packages:
            pkgs = self.getPackages(name)
            if not pkgs:
                log.logError(
                    f"Unable to locate package {name} for"
                    f" {self.config.os}/{self.config.os_version} (Release:"
                    f" {self.config.release})"
                )
                continue

            try:
                pkg = pkgs[0]
                try:
                    if package.download(pkg, self.config):
                        downloaded.append(name)

                    continue
                except Exception as err:
                    log.logError(f"{err}")
                    return False

            except IndexError as err:
                log.logError(f"{err}")
                return False

        return downloaded

    def latestPackage(self, name):
        packages = []
        for pkg in sorted(self.getPackages(name)):
            cert = utils.readCertificate(name, self.config)
            if not cert:
                return False
            if not cert.updatesUntil or pkg.version <= cert.updatesUntil:
                packages.append(pkg.version)

        sorted(packages, reverse=True)
        if packages:
            return packages[0]

        return None

    def getPackagePath(self, pkg):
        if pkg.filePath:
            return os.path.join(
                EXTEND_DIR, os.path.join(os.path.split(pkg.filePath)[0], pkg.name)
            )

        if pkg.type == PT.CONFIG:
            return os.path.join(EXTEND_DIR, pkg.name)

        if pkg.arch == NO_ARCH:
            return self.config.noArchDir

        return self.config.archDir
