############################################################################
# Copyright (C) 2019 by gempa GmbH                                         #
#                                                                          #
# All Rights Reserved.                                                     #
#                                                                          #
# NOTICE: All information contained herein is, and remains                 #
# the property of gempa GmbH and its suppliers, if any. The intellectual   #
# and technical concepts contained herein are proprietary to gempa GmbH    #
# and its suppliers.                                                       #
# Dissemination of this information or reproduction of this material       #
# is strictly forbidden unless prior written permission is obtained        #
# from gempa GmbH.                                                         #
#                                                                          #
#  Author: Enrico Ellguth, Ralph Henneberger                               #
#  Email: enrico.ellguth@gempa.de, ralix@gempa.de                          #
############################################################################

import configparser
import os
import sys

from http import HTTPStatus

import subprocess

from libs import utils, log

from libs.const import (
    BASE_URL,
    LICENSES,
    SC,
    SC3,
    urljoin,
    BASE_PUBLIC_URL,
    LICENSE_INFO_DAYS,
    FILE_ENCODING,
    EXTEND_DIR,
    ANY_NOARCH_DIR,
    CORE_DB,
    DEFAULT_RELEASE,
    DEFAULT_ARCH,
)
from libs.utils import checkInput, getInput

CONFIG_FILE = "gsm.conf"
SECTION = "Settings"


def getConfigParser():
    return configparser.RawConfigParser(allow_no_value=True)


class Config(object):
    @staticmethod
    def IsSC3(release):
        return release.startswith("jakarta-")

    def isSC3(self):
        return self.IsSC3(self.release)

    @staticmethod
    def InstallPrefix(release):
        return SC3 if Config.IsSC3(release) else SC

    def installPrefix(self):
        return self.InstallPrefix(self.release)

    @staticmethod
    def expandAbsPath(path):
        return os.path.abspath(os.path.expanduser(path))

    def readInstallInfoFile(self):
        parser = getConfigParser()
        if not parser.read(os.path.join(self.localDir, self.internalConfig)):
            log.logError(f"Could not parse config file: {self.internalConfig}")
            return None

        return parser

    def writeInstallInfoFile(self, release, installPath, gsmBaseDir, dataDir):
        if not os.path.isdir(self.localDir):
            log.logError(
                f"The local_dir {(self.localDir)} defined in gsm.conf does not exist."
            )
            return False
        try:
            parser = getConfigParser()
            parser.add_section("CONFIG")
            parser.set("CONFIG", "release", release)
            parser.set("CONFIG", "install_path", installPath)
            parser.set("CONFIG", "baseDir", gsmBaseDir)
            parser.set("CONFIG", "data_dir", dataDir)
            parser.set("CONFIG", "config", "0.0")
            self.write(parser, self.internalConfig)
        except IOError as e:
            log.logError(f"Could not write {self.internalConfig} file: {e}")
            return False

        return True

    def updateInstallInfoFile(self, section, value):
        try:
            parser = getConfigParser()
            parser.read(self.internalConfig)
            parser.set("CONFIG", section, value)
            self.write(parser, self.internalConfig)
        except IOError as e:
            log.logError(f"Could not write {self.internalConfig} file: {e}")
            return False

        return True

    def writeInitialConfigFile(self, release, installPath, gsmBaseDir, dataDir):
        try:
            if not os.path.isfile(self.internalConfig):
                self.writeInstallInfoFile(release, installPath, gsmBaseDir, dataDir)
                self.updateInstallInfoFile("config", self.getLatestMigrationVersion())
        except IOError as e:
            log.logError(f"Could not write {self.internalConfig} file: {e}")
            return False

        return True

    def getLatestMigrationVersion(self):
        path = "migrations"
        versions = []

        fileNames = os.listdir(path)
        fileNames.sort()

        for fileName in fileNames:
            toks = fileName.split("_to_")
            if len(toks) != 2:
                continue

            begin = toks[0]
            end = toks[1]
            if begin < "0.0":
                continue

        versions.append(end)
        return versions[-1]

    def getCurrentVersion(self):
        if os.path.isfile(self.internalConfig):
            parser = self.readInstallInfoFile()
            try:
                value = parser.get("CONFIG", "config")

            except Exception:
                value = "0.0"

        return value

    def getMigrationFiles(self):
        currentVersion = self.getCurrentVersion()
        files = []
        try:
            fileNames = os.listdir("migrations")
        except FileNotFoundError:
            log.logError(
                f"The {os.path.join(self.baseDir, 'migrations')} does not exist."
            )
            return ["failed"]
        fileNames.sort()
        end = None

        for fileName in fileNames:
            toks = fileName.split("_to_")
            if len(toks) != 2:
                continue

            begin = toks[0]
            end = toks[1]
            if begin < currentVersion:
                continue

            files.append([fileName, end])

        return files

    def makeMigration(self):
        files = self.getMigrationFiles()
        if not files:
            print("The configuration is up to date.")
            return True
        if files[0] == "failed":
            return False

        absEnd = None

        for fileName, end in files:
            print(f"Running Migration step {fileName}".replace("_to_", " --> "))

            if subprocess.call(f"migrations/{fileName}", shell=True):
                print("The the migration was not successful.")
                return False

            absEnd = end

        return self.updateInstallInfoFile("config", absEnd)

    def checkMigration(self):
        migrationCount = len(self.getMigrationFiles())
        if migrationCount:
            print(
                f"Found {migrationCount} outstanding migration(s). "
                "It is required to apply them first."
                "\nRun: ./gsm migrate"
            )
            return False
        return True

    def checkInstallInfoWithExistConfig(self):
        if not os.path.isfile(self.internalConfig):
            return False
        parser = self.readInstallInfoFile()
        if not parser:
            self.writeInstallInfoFile(
                self.release, self.installPath, self.baseDir, self.dataDir
            )

        cfgRelease = parser.get("CONFIG", "release").lower()
        cfgInstallPath = parser.get("CONFIG", "install_path")
        try:
            cfgBaseDir = parser.get("CONFIG", "baseDir")
        except configparser.NoOptionError:
            self.updateInstallInfoFile("baseDir", self.baseDir)
            cfgBaseDir = self.baseDir
        try:
            cfgDataPath = parser.get("CONFIG", "data_dir")
        except configparser.NoOptionError:
            self.updateInstallInfoFile("data_dir", self.dataDir)
            cfgDataPath = self.dataDir

        failed = True

        if not os.getcwd() == cfgBaseDir:
            q = (
                "Note: gsm directory has been copied/moved to a new location. "
                "This may lead to undefined behavior!\n\n"
                "Do you want fix it?"
            )
            if not checkInput(q, "Y", "Yn"):
                sys.exit(0)

            try:
                print("")
                self.setDownloadDir(
                    getInput("Download dir", os.path.join(self.baseDir, "packages"))
                )
                self.setSyncDir(
                    getInput("Sync dir", os.path.join(self.baseDir, "sync"))
                )
                self.setLocalDir(
                    getInput("Local dir", os.path.join(self.baseDir, "local"))
                )
                self.internalConfig = os.path.join(self.localDir, ".config")
                print("")

                self.updateInstallInfoFile("baseDir", self.baseDir)

                self.update({"download_dir": self.downloadDir}, [])
                self.update({"local_dir": self.localDir}, [])
                self.update({"sync_dir": self.syncDir}, [])

            except KeyboardInterrupt:
                print("\n\nSetup aborted by user.")
            sys.exit(0)

        if not cfgRelease == self.release:
            log.logWarning(
                f"The SeisComP release has been changed from {cfgRelease} to"
                f" {self.release} in the gsm.conf file. This can lead to problems"
                " because possible migration steps were not executed. It is strongly"
                f" recommended to change the release back to {cfgRelease} and then do the"
                " release upgrade with './gsm do-release-upgrade'."
            )
            failed = False

        if not cfgInstallPath == self.installPath:
            q = (
                "The SeisComP installation path was manually changed from"
                f" {cfgInstallPath} to {self.installPath}.\n"
                "Do you want fix it?"
            )

            if not checkInput(q, "Y", "Yn"):
                sys.exit(0)
            self.updateInstallInfoFile("install_path", self.installPath)

        if not cfgDataPath == self.dataDir:
            log.logWarning(
                "The data dir installation path was manually changed from"
                f" {cfgDataPath} to {self.dataDir}."
            )
            failed = False

        return failed

    def checkInstallInfo(self):
        if os.path.isfile(self.internalConfig):
            failed = self.checkInstallInfoWithExistConfig()

            if not failed:
                if not checkInput("\nDo you want to continue?", "y", "yn"):
                    return False

        else:
            return self.writeInstallInfoFile(
                self.release, self.installPath, self.baseDir, self.dataDir
            )

        return True

    def __init__(self, baseDir):
        # directories and files relative to gsm base directory
        self.baseDir = self.expandAbsPath(baseDir)
        self.fileName = os.path.join(self.baseDir, CONFIG_FILE)

        # configuration parameters
        self.os = ""
        self.os_version = ""
        self.arch = DEFAULT_ARCH
        self.baseURL = ""
        self.user = ""
        self.password = ""
        self.release = DEFAULT_RELEASE

        # will be initialized in subsequent set methods
        self.downloadDir = None
        self.installPath = None
        self.installDir = None
        self.rootDir = None
        self.licenseDir = None
        self.libDir = None
        self.userConfigDir = None

        self.homeDir = self.expandAbsPath("~")

        self.setDownloadDir(os.path.join(self.baseDir, "packages"))
        self.setInstallDir(self.homeDir)

        self.verify = True
        self.debug = False
        self.dataDir = "/home/data"

        # modified by lib/synccache ???
        self.dbUrl = urljoin(BASE_URL, "coredb")
        self.dbFileName = CORE_DB

        self.syncDir = os.path.join(self.baseDir, "sync")
        self.localDir = os.path.join(self.baseDir, "local")
        self.internalConfig = os.path.join(self.localDir, ".config")

        # command line only parameters
        self.showAll = False
        self.downloadOnly = False
        self.showUrls = False
        self.assumeYes = False
        self.offline = False

        self.licenseInfoDays = None

        self.forceDownload = False

        self.archDir = None
        self.noArchDir = None
        self.extendDir = None

        self.maxlenPackageFullName = 0

        self.updateOrDowngrade = False
        self.fromUpdate = False

    def write(self, parser, fileName):
        tmpFile = f"{fileName}.tmp"
        with open(tmpFile, "w", encoding=FILE_ENCODING) as f:
            parser.write(f)
        os.replace(tmpFile, fileName)

    # Update and/or remove values from configuration
    # @param updates Map of new or to be updates key value pairs
    # @param removals List of keys to remove
    def update(self, updates, removals, parser=None):
        if not parser:
            parser = getConfigParser()
            parser.read(self.fileName)

        for key, value in updates.items():
            parser.set(SECTION, key, value)
        for key in removals:
            parser.remove_option(SECTION, key)

        self.write(parser, self.fileName)

    def read(self):
        parser = getConfigParser()
        updates = {}
        removals = []

        if not parser.read(self.fileName):
            return (
                f"{log.getLogPrefix('ERROR')}Could not parse config file:"
                f" {self.fileName}"
            )

        try:
            self.os = parser.get(SECTION, "os").lower()
            if "centos" in self.os.lower():
                self.os = "rhel"
                updates["os"] = self.os

        except configparser.NoOptionError:
            pass

        try:
            self.os_version = parser.get(SECTION, "os_version")
        except configparser.NoOptionError:
            pass

        try:
            self.arch = parser.get(SECTION, "arch").lower()
        except configparser.NoOptionError:
            pass

        try:
            self.baseURL = parser.get(SECTION, "base_url")
        except configparser.NoOptionError:
            pass

        try:
            self.user = parser.get(SECTION, "user")
        except configparser.NoOptionError:
            pass

        try:
            self.password = parser.get(SECTION, "password")
        except configparser.NoOptionError:
            pass

        try:
            self.release = parser.get(SECTION, "release").lower()
        except configparser.NoOptionError:
            pass

        try:
            self.setDownloadDir(parser.get(SECTION, "download_dir"))
        except configparser.NoOptionError:
            pass

        try:
            self.setSyncDir(parser.get(SECTION, "sync_dir"))
        except configparser.NoOptionError:
            pass

        try:
            self.setLocalDir(parser.get(SECTION, "local_dir"))
        except configparser.NoOptionError:
            pass

        try:
            # new configuration variable install_path which is a combination of
            # the installDir and rootDir, e.g. /home/sysop/seiscomp
            self.setInstallPath(parser.get(SECTION, "install_path"))
        except configparser.NoOptionError:
            # old configuration variable install_dir not including the
            # rootDir, e.g. /home/sysop
            try:
                self.setInstallDir(parser.get(SECTION, "install_dir"))
                print(
                    "Found deprecated option 'install_dir' which was replaced by"
                    " 'install_path'. While the 'install_dir' named the base path"
                    " without the seiscomp root folder, the 'install_path' will"
                    " include it. Based on your current settings we suggest to set\n "
                    f" 'install_path = {self.installPath}"
                )
                q = "Replace install_dir with install_path?"
                if checkInput(q, "y", "yn"):
                    updates["install_path"] = self.installPath
                    removals.append("install_dir")

            except configparser.NoOptionError:
                return (
                    f"{log.getLogPrefix('ERROR')}Could not find mandatory config"
                    " option: install_path"
                )

        try:
            self.setDataDir(parser.get(SECTION, "data_dir"))
        except configparser.NoOptionError:
            pass

        try:
            self.verify = parser.getboolean(SECTION, "verify")
        except configparser.NoOptionError:
            pass

        try:
            self.debug = parser.getboolean(SECTION, "debug")
        except configparser.NoOptionError:
            pass

        try:
            self.dbUrl = parser.get(SECTION, "db_url")
        except configparser.NoOptionError:
            pass

        try:
            self.dbFileName = parser.get(SECTION, "db_filename")
        except configparser.NoOptionError:
            pass

        try:
            self.licenseInfoDays = parser.get(SECTION, "license_info_days")
        except configparser.NoOptionError:
            pass

        self.archDir = os.path.join(self.release, self.os, self.os_version, self.arch)

        self.noArchDir = os.path.join(self.release, ANY_NOARCH_DIR)

        self.extendDir = os.path.join(self.syncDir, EXTEND_DIR)

        if updates or removals:
            self.update(updates, removals, parser)

        return None

    def setNoArchDir(self, noArchDir):
        self.noArchDir = noArchDir

    def setArchDir(self, archDir):
        self.archDir = archDir

    def setDataDir(self, dataDir):
        self.dataDir = self.expandAbsPath(dataDir)

    def setLicenseInfoDays(self, licenseInfoDays):
        self.licenseInfoDays = licenseInfoDays

    def setDownloadDir(self, downloadDir):
        self.downloadDir = self.expandAbsPath(downloadDir)
        self.downloadLicenseDir = os.path.join(self.downloadDir, LICENSES)

    def setSyncDir(self, syncDir):
        self.syncDir = self.expandAbsPath(syncDir)

    def setLocalDir(self, localDir):
        self.localDir = self.expandAbsPath(localDir)
        self.internalConfig = os.path.join(self.localDir, ".config")

    def setInstallDir(self, installDir):
        self.installDir = self.expandAbsPath(installDir)
        self.rootDir = self.installPrefix()
        self.installPath = os.path.join(self.installDir, self.rootDir)
        self.updatePaths()

    def setInstallPath(self, installPath):
        self.installPath = self.expandAbsPath(installPath)
        self.installDir = os.path.dirname(self.installPath)
        self.rootDir = os.path.basename(self.installPath)
        # This is a security precaution
        if not self.rootDir:
            raise Exception("Root directory not supported as install path")

        self.updatePaths()

    def updatePaths(self):
        self.licenseDir = os.path.join(self.installPath, "share", LICENSES)
        self.libDir = os.path.join(self.installPath, "lib")
        self.userConfigDir = os.path.join(self.homeDir, f".{self.installPrefix()}")

    def setLoginCredentials(self):
        self.user = getInput("Username", self.user)
        self.password = getInput("Password", self.password)

        # guess default baseURL based on user name
        if not self.baseURL:
            self.baseURL = urljoin(BASE_URL, "packages", self.user.upper())

        url = getInput("Repository URL", self.baseURL)

        # if the URL does not contain a slash assume that only the last
        # path component was entered
        self.baseURL = url if "/" in url else f"{self.baseURL.rpartition('/')[0]}/{url}"

    def setup(self):
        print(
            """====================================================================
Package manager setup
====================================================================

This initializes the configuration of your package manager.
If you already made adjustments to the configuration file then your
current settings will be suggested as default values and may be
confirmed simply by pressing enter.
The configuration file is only written after the setup is completed
entirely.
"""
        )

        # detect operating
        if not all((self.os, self.os_version, self.arch)):
            os_, version, arch = utils.detectOS()
        if not self.os:
            self.os = os_
        if not self.os_version:
            self.os_version = version
        if not self.arch:
            self.arch = arch

        try:
            self.os = getInput("Operating system", self.os)
            self.os_version = getInput("Operating system version", self.os_version)
            self.arch = getInput("System architecture", self.arch)

            if checkInput(
                "Do you want to use the public SeisComP repository without"
                " gempa modules?",
                "n",
                "yn",
            ):

                self.setLoginCredentials()

                ret, retCode = utils.checkUrlAccess(
                    self.baseURL, self.user, self.password
                )
                while not ret:
                    if retCode == 401:
                        print("Invalid login credentials.")
                    elif retCode == 404:
                        print("Invalid repository URL.")
                    else:
                        print(f"HTTP Error: {HTTPStatus(retCode).phrase} ")

                    self.setLoginCredentials()
                    ret, retCode = utils.checkUrlAccess(
                        self.baseURL, self.user, self.password
                    )
            else:
                self.baseURL = BASE_PUBLIC_URL

            self.release = getInput("SeisComP release", self.release)
            while not utils.isReleaseAvailable(self.release):
                print(
                    f"Unknown SeisComP release '{self.release}' or release no longer available!"
                )
                self.release = getInput("SeisComP release", self.release)

            self.setDownloadDir(getInput("Download dir", self.downloadDir))
            self.setInstallPath(getInput("Install path", self.installPath))
            self.setDataDir(getInput("Data path", self.dataDir))
            if not self.licenseInfoDays:
                self.licenseInfoDays = LICENSE_INFO_DAYS
            self.setLicenseInfoDays(
                getInput(
                    "Number of days before "
                    "the license expires message "
                    "is displayed",
                    self.licenseInfoDays,
                )
            )
        except KeyboardInterrupt:
            print("\n\nSetup aborted by user.")
            return False

        parser = getConfigParser()

        parser.add_section(SECTION)
        parser.set(SECTION, "os", self.os)
        parser.set(SECTION, "os_version", self.os_version)
        parser.set(SECTION, "arch", self.arch)
        parser.set(SECTION, "base_url", self.baseURL)
        parser.set(SECTION, "user", self.user)
        parser.set(SECTION, "password", self.password)
        parser.set(SECTION, "release", self.release)
        parser.set(SECTION, "download_dir", self.downloadDir)
        parser.set(SECTION, "install_path", os.path.join(self.installDir, self.rootDir))
        parser.set(SECTION, "data_dir", self.dataDir)
        parser.set(SECTION, "verify", self.verify)
        parser.set(SECTION, "debug", self.debug)
        parser.set(SECTION, "sync_dir", self.syncDir)
        parser.set(SECTION, "local_dir", self.localDir)
        parser.set(SECTION, "license_info_days", self.licenseInfoDays)

        self.write(parser, self.fileName)

        self.writeInitialConfigFile(
            self.release,
            os.path.join(self.installDir, self.rootDir),
            self.baseDir,
            self.dataDir,
        )

        return True
