############################################################################
# Copyright (C) 2019 by gempa GmbH                                         #
#                                                                          #
# All Rights Reserved.                                                     #
#                                                                          #
# NOTICE: All information contained herein is, and remains                 #
# the property of gempa GmbH and its suppliers, if any. The intellectual   #
# and technical concepts contained herein are proprietary to gempa GmbH    #
# and its suppliers.                                                       #
# Dissemination of this information or reproduction of this material       #
# is strictly forbidden unless prior written permission is obtained        #
# from gempa GmbH.                                                         #
#                                                                          #
#  Author: Enrico Ellguth, Ralph Henneberger                               #
#  Email: enrico.ellguth@gempa.de, ralix@gempa.de                          #
############################################################################

import os

from tqdm import tqdm

from libs import log, utils
from libs.cache import synccache

from libs.const import SC, PT
from libs.utils import checkInput, getInput, detectOS

from libs.const import REMOVE, REMOVE_OK, ANY_NOARCH_DIR


class OSUpgrade(object):
    def __init__(self, config, localCache):
        self.config = config
        self.localCache = localCache
        self.syncCache = None
        self.installedPackages = None
        self.new_os = None
        self.new_os_version = None
        self.new_arch = None

        self.old_os = config.os
        self.old_os_version = config.os_version
        self.old_arch = config.arch

    def prepareUpgrade(self):
        detect_os_name, detect_os_version, detect_os_arch = detectOS()
        count = 1

        print(
            "\nIf your SeisComP installation is not under version control, we"
            " recommend to make a backup before continuing."
        )
        if checkInput("Do you want to continue?", "n", "yn"):
            return False

        os_names = sorted(
            os.listdir(os.path.join(self.config.syncDir, SC, self.config.release))
        )

        if len(os_names) == 1:
            self.new_os = os_names[0]
        else:
            print("")
            print("Available operating systems:")
            print("")

            for os_name in os_names:
                print(f"{count}. {os_name}")
                count += 1

            print("")

            try:
                idx = getInput(f"Upgrade to Option ({detect_os_name})")
                if not idx:
                    idx = os_names.index(detect_os_name) + 1
                value = int(idx)
                self.new_os = os_names[value - 1]
            except KeyboardInterrupt:
                return False

            except Exception:
                print("Invalid entry. Please select an entry from the list.")
                return False

        count = 1
        os_versions = sorted(
            os.listdir(
                os.path.join(self.config.syncDir, SC, self.config.release, self.new_os)
            )
        )

        if len(os_versions) == 1:
            self.new_os_version = os_versions[0]
        else:
            print("")
            print("Available Versions:")
            print("")

            for os_version in os_versions:
                print(f"{count}. {os_version}")
                count += 1
            print("")

            try:
                if self.new_os == detect_os_name:
                    idx = getInput(f"Upgrade to Option ({detect_os_version})")
                else:
                    idx = getInput("Upgrade to Option")

                if not idx and self.new_os == detect_os_name:
                    idx = os_versions.index(detect_os_version) + 1
                value = int(idx)
                self.new_os_version = os_versions[value - 1]
            except KeyboardInterrupt:
                return False

            except Exception:
                print("Invalid entry. Please select an entry from the list.")
                return False

        count = 1
        os_archs = os.listdir(
            os.path.join(
                self.config.syncDir,
                SC,
                self.config.release,
                self.new_os,
                self.new_os_version,
            )
        )

        if len(os_archs) == 1:
            self.new_arch = os_archs[0]
        else:
            print("Available Architectures:")
            print("")

            for os_arch in os_archs:
                print(f"{count}. {os_arch}")
                count += 1
            print("")

            try:
                idx = getInput(f"Upgrade to Option ({detect_os_arch})")
                if not idx:
                    idx = os_archs.index(detect_os_arch) + 1
                value = int(idx)
                self.new_arch = os_archs[value - 1]
            except KeyboardInterrupt:
                return False

            except Exception:
                print("Invalid entry. Please select an entry from the list.")
                return False

        if (
            self.config.os == self.new_os
            and self.config.os_version == self.new_os_version
            and self.config.arch == self.new_arch
        ):

            print("")
            print("Nothing todo ...")

            return False

        if self.new_os and self.new_os_version and self.new_arch:

            self.config.update({"os": self.new_os}, [])
            self.config.update({"os_version": self.new_os_version}, [])
            self.config.update({"arch": self.new_arch}, [])

            self.config.os = self.new_os
            self.config.os_version = self.new_os_version
            self.config.arch = self.new_arch

            packages = list(self.localCache.getIndex().keys())
            self.installedPackages = []

            for name in packages:
                pkg = self.localCache.getPackage(name)
                if pkg.type == PT.APP:
                    self.installedPackages.append(pkg.name)

            if not self.findObsoletePackages():
                return False

            if not self.downloadNewPackages():
                return False

            stopUpgrade = False
            for name in self.installedPackages:
                if not self.localCache.isInstallable(self.localCache.getPackage(name)):
                    stopUpgrade = True

            if stopUpgrade:
                log.logError(
                    "Could not perform the upgrade because at least one application is"
                    " not covered by the license. Please get in touch with"
                    " info@gempa.de"
                )

                return False

        return True

    def removePackages(self):
        members = utils.sorting(self.installedPackages)
        count = 0
        progress = tqdm(
            members,
            ascii=True,
            bar_format=utils.getTerminalSizeBarFormat(self.config),
        )
        for member in progress:
            pkgName, _ = utils.getPackageName(self.localCache.getPackage(member))
            progress.set_description(
                f"{log.getLogPrefix(REMOVE)}{pkgName: <{self.config.maxlenPackageFullName:}}"
            )
            count = count + 1
            done = f"DONE ({len(progress)}/{count})"
            self.localCache.removePackage(member, False)
            if count == len(members):
                progress.set_description(
                    f"{log.getLogPrefix(REMOVE_OK)}"
                    f"{done: <{self.config.maxlenPackageFullName:}}"
                )

        return True

    def downloadNewPackages(self):
        downloadPackages = []

        self.config.setArchDir(
            os.path.join(
                self.config.release,
                self.config.os,
                self.config.os_version,
                self.config.arch,
            )
        )
        self.config.setNoArchDir(os.path.join(self.config.release, ANY_NOARCH_DIR))

        syncCache = synccache.SyncCache(self.config)

        if not self.installedPackages:
            print("\nNo installed packages found.")
        for name in self.installedPackages:

            if syncCache.contains([name], False):
                downloadPackages.append(name)

        utils.setMaxlenPackageFullName(
            downloadPackages, self.localCache, self.syncCache, self.config
        )

        if not self.config.offline:
            utils.checkUrlAccessProgress(self.config, syncCache, downloadPackages)

            ret = syncCache.downloadPackages(downloadPackages)
        else:
            ret = True
        failed = []
        fileNotExist = []
        if ret:
            for name in downloadPackages:
                pkgSync = syncCache.getPackages(name)[0]
                fileName = os.path.join(self.config.downloadDir, pkgSync.fileName)
                if os.path.exists(fileName):
                    fail = utils.hashfile(
                        self.config, fileName, pkgSync.checksum, pkgSync
                    )
                    if not fail:
                        failed.append(pkgSync.name)
                        if os.path.isfile(fileName):
                            os.remove(fileName)
                else:
                    fileNotExist.append(pkgSync)

        if fileNotExist:
            for pkg in fileNotExist:
                print(
                    f"The {pkg.fileName} file does not exist in the"
                    f" {self.config.downloadDir} directory."
                )
                return False

        if failed:
            print(
                "\nChecksum validation for the following packages failed:"
                f" {', '.join(failed)}"
            )
            print("\nTry to run ./gsm update to refresh the package index.")

            return False

        return ret

    def findObsoletePackages(self):
        config = self.config
        self.syncCache = synccache.SyncCache(config)
        if not self.syncCache.getIndex():
            return False

        obsoletePackages = []
        for name in self.installedPackages:
            if not self.syncCache.contains([name], False):
                obsoletePackages.append(name)

        if obsoletePackages:
            log.logInfo(
                "The following package(s) are no longer available in the new os"
                f" ({self.new_os} {self.new_os_version} ({self.new_arch})):"
                f" {', '.join(obsoletePackages)}"
            )
        if checkInput("Do you want to continue?", "n", "yn"):
            return False
        print("")
        return True

    def installPackages(self, app):
        installPackages = []
        app.init()

        self.config.assumeYes = True
        for name in self.installedPackages:
            if self.syncCache.contains([name], False):
                installPackages.append(name)

        return app.install(installPackages, True)

    def run(self, app):
        self.config.updateOrDowngrade = True
        if not self.prepareUpgrade():
            return False
        if not self.config.downloadOnly:
            if not self.removePackages():
                return False

            return self.installPackages(app)

        return True
