############################################################################
# Copyright (C) 2019 by gempa GmbH                                         #
#                                                                          #
# All Rights Reserved.                                                     #
#                                                                          #
# NOTICE: All information contained herein is, and remains                 #
# the property of gempa GmbH and its suppliers, if any. The intellectual   #
# and technical concepts contained herein are proprietary to gempa GmbH    #
# and its suppliers.                                                       #
# Dissemination of this information or reproduction of this material       #
# is strictly forbidden unless prior written permission is obtained        #
# from gempa GmbH.                                                         #
#                                                                          #
#  Author: Enrico Ellguth, Ralph Henneberger                               #
#  Email: enrico.ellguth@gempa.de, ralix@gempa.de                          #
############################################################################

import glob
import logging
import os
import tarfile
import sys

from datetime import datetime

from shutil import copy

from tqdm import tqdm

from libs import pkgbuild, utils, log
from libs.const import (
    DOC_URL,
    NIGHTLY,
    SC,
    SC3,
    PT,
    GSM_INFO_URL,
    urljoin,
    FILE_ENCODING,
    DEFAULT_TIME_FORMAT,
    DEFAULT_UPDATE_DATE_FORMAT,
    CRT_EXTENSION,
    LICENSE_TXT_EXTENSION,
    CHANGELOG,
    INSTALL_OK,
    CHECKING_OK,
    LICENSE_MAPPING,
    PUBLIC_PACKAGES,
    BASE_PUBLIC_URL,
)

from libs.utils import checkInput

FORMAT = "0.1"


class Package(object):
    def __init__(self, name, path):
        self.name = name
        self.directory = path

        self.desc = ""
        self.keywords = ""
        self.fileName = ""
        self.installSize = 0
        self.checksum = ""
        self.version = ""
        self.arch = ""
        self.buildDate = ""
        self.installPath = None
        self.dscFile = ""
        self.downloadSize = 0
        self.installDate = ""
        self.license = ""
        self.release = ""
        self.fileSize = 0
        self.format = FORMAT
        self.type = PT.APP
        self.conflicts = None
        self.filePath = None
        self.changeLog = None
        self.os = ""
        self.osVersion = ""
        self.licenseFile = None
        self.installScript = None
        self.optional = None

    @classmethod
    def fromFile(cls, file, path):
        def getValue(key):
            v = symbols.get(key)
            if isinstance(v, list):
                return " ".join(v)

            return v

        def getOptValue(key):
            return getValue(key) if key in symbols else None

        # read description file
        symbols = pkgbuild.readPkgsDescFile(file)

        # create instance
        pkg = cls(getValue("pkgname"), path)
        pkg.dscFile = file

        # mandatory values
        pkg.desc = getValue("pkgdesc")
        pkg.fileName = getValue("filename")
        pkg.downloadSize = getValue("pkgsize")
        pkg.fileSize = getValue("pkginstallsize")

        pkg.checksum = getOptValue("sha256") or getValue("sha512sums")

        pkg.version = getValue("pkgver")

        pkg.os = getValue("pkgos")
        pkg.osVersion = getValue("pkgosversion")
        pkg.arch = getValue("arch")

        pkg.license = getValue("license")
        pkg.buildDate = getValue("builddate")
        pkg.installSize = getValue("pkginstallsize")

        # optional values
        pkg.release = getOptValue("pkgrel") or pkg.release
        pkg.installPath = getOptValue("pkginstallpath")
        pkg.installDate = getOptValue("installdate") or pkg.installDate
        pkg.format = getOptValue("pkgfileversion") or pkg.format
        pkg.keywords = getOptValue("pkgkeywords") or pkg.keywords
        pkg.changeLog = getOptValue("pkgchangelog") or pkg.changeLog
        pkg.conflicts = getOptValue("conflicts") or pkg.conflicts
        pkg.optional = getOptValue("optional") or pkg.optional
        pkg.filePath = getOptValue("pkgfilepath") or pkg.filePath
        pkg.licenseFile = getOptValue("pkglicensefile") or pkg.licenseFile
        if pkg.licenseFile == "None":
            pkg.licenseFile = None
        pkg.installScript = getOptValue("pkginstallscript") or pkg.installScript

        # special case type:
        # currently set only for PT.DATA and PT.CONFIG
        # if empty use PT.APP unless package name is gsm
        pt = getOptValue("pkgtype")
        if pt:
            pkg.type = pt
        elif pkg.name == PT.GSM:
            pkg.type = PT.GSM

        return pkg

    def __lt__(self, other):
        return self.version > other.version

    def fullName(self):
        return self.name + "=" + self.version

    def showInfo(self, showInstallDate):
        if self.release:
            print(f"Release          : {self.release}")

        print(f"Version          : {self.version}")

        if not self.filePath and self.type != PT.CONFIG:
            urlStr = urljoin(DOC_URL, f"{self.name}/current/")
            print(f"Documentation    : {urlStr}")

        print(f"Architecture     : {self.arch}")
        print(f"Licenses         : {self.license}")
        print(f"Description      : {self.desc}")
        print(f"Filename         : {self.fileName}")
        print(f"Download size    : {utils.humanbytes(self.downloadSize)}")
        print(f"Installed size   : {utils.humanbytes(self.installSize)}")
        print(f"SHA256           : {self.checksum}")
        print(
            "Build date       :"
            f" {self.buildDate}{'' if self.buildDate == 'N/A' else ' UTC'}"
        )
        print(f"Package type     : {self.type}")
        if self.optional:
            print(f"Package optional : {self.optional}")

        if showInstallDate:
            print(
                "Install date     :"
                f" {self.installDate}{'' if self.installDate == 'N/A' else ' UTC'}"
            )

    def installLicense(self, config):
        if not utils.createDirectory(config.licenseDir):
            return False

        licenseFileName, members = getLicenseFiles(self, config.downloadLicenseDir)
        progress = tqdm(
            members,
            ascii=True,
            bar_format=utils.getTerminalSizeBarFormat(config),
        )

        licenseFileName = licenseFileName + " license"
        progress.set_description(
            f"{log.getLogPrefix('INSTALL')}{licenseFileName: <{config.maxlenPackageFullName:}}"
        )
        for member in progress:
            try:
                copy(member, config.licenseDir)
            except BaseException as e:
                progress.set_description(
                    f"{log.getLogPrefix(INSTALL_OK)}{licenseFileName} license , {e}"
                )
                return False
            progress.set_description(
                f"{log.getLogPrefix(INSTALL_OK)}"
                f"{licenseFileName: <{config.maxlenPackageFullName:}}"
            )
        logging.info("license for %s was installed", self.name)

        return True


def getLicenseFiles(pkg, downloadLicenseDir):
    licenseFileName = getLicenseFileName(pkg)
    baseName = os.path.join(downloadLicenseDir, licenseFileName)

    return licenseFileName, glob.glob(rf"{baseName}*")


def getLicenseFileName(pkg):
    if pkg.licenseFile:
        return pkg.licenseFile

    if pkg.name in LICENSE_MAPPING:
        return LICENSE_MAPPING.get(pkg.name)

    return pkg.name


def getPackagesUpdateInfo(syncCache, localCache, config):
    updatablePackages = []
    packages = []
    for pkgName in sorted(syncCache.getIndex()):
        if pkgName == "gsm":
            continue

        pkg = localCache.getPackage(pkgName)
        if pkg is None:
            continue

        installCandidate = syncCache.getPackages(pkg.name)

        if config.release == NIGHTLY:
            if pkg.type == PT.APP:
                updatablePackages.append((pkg.name, installCandidate[0]))
                if not localCache.isHold(pkg.name):
                    packages.append(pkg.name)
            if not pkg.type == PT.APP:
                if not pkg.version == installCandidate[0].version:
                    updatablePackages.append((pkg.name, installCandidate[0]))
                    if not localCache.isHold(pkg.name):
                        packages.append(pkg.name)

        elif installCandidate and (not pkg.version == installCandidate[0].version):
            updatablePackages.append((pkg.name, installCandidate[0]))
            if not localCache.isHold(pkg.name):
                packages.append(pkg.name)

        elif (
            installCandidate
            and (pkg.version == installCandidate[0].version)
            and not pkg.checksum == installCandidate[0].checksum
            and localCache.isInstalledToday(pkg.installDate)
        ):
            updatablePackages.append((pkg.name, installCandidate[0]))
            if not localCache.isHold(pkg.name):
                packages.append(pkg.name)

    if not updatablePackages:
        print("\nNothing to do.")
        return False, None

    if config.release == NIGHTLY:
        print(
            "\nNightly release detected. It is strongly recommended to update all"
            " packages!\n"
        )
    else:
        print("\nThe following packages can be updated:\n")

    for name, pkgs in updatablePackages:
        cert = None
        pkg = localCache.getPackage(name)
        if pkg.license != "None":

            licenseFileName = getLicenseFileName(pkg)

            cert = utils.readCertificate(licenseFileName, config)
            if cert and cert.updatesUntil and cert.updatesUntil < pkg.version:
                print(
                    f"{pkg.name} ({pkg.version}) --> {pkgs.version} (Unsupported by"
                    f" license. Updates until: {cert.updatesUntil})"
                )
                packages.remove(name)
                continue

        if config.licenseInfoDays and cert:
            lincensInfo = utils.prettyFormatLicenseInfo(cert, config.licenseInfoDays)
        else:
            lincensInfo = ""

        if localCache.isHold(pkg.name):
            holdInfo = "(will not be updated, it's on hold)"
        else:
            holdInfo = ""

        if pkg.version == pkgs.version:
            print(
                f"{pkg.name} ({pkg.version}) --> {pkgs.version} (force update) {lincensInfo}"
                f"{holdInfo}"
            )
        else:
            print(
                f"{pkg.name} ({pkg.version}) --> {pkgs.version} {lincensInfo}"
                f"{holdInfo}"
            )

    if not packages:
        return False, None

    return True, packages


def download(pkg, config):
    if not utils.createDirectory(config.downloadDir):
        return False

    if not os.access(config.downloadDir, os.W_OK):
        log.logError(f"Permission denied on directory {config.downloadDir}")
        sys.exit(0)

    if pkg.type == PT.CONFIG:
        url = urljoin(config.baseURL, pkg.name, pkg.fileName)
    elif pkg.filePath:
        url = urljoin(
            config.baseURL,
            os.path.join(os.path.split(pkg.filePath)[0], pkg.name),
            pkg.fileName,
        )
    elif pkg.name in PUBLIC_PACKAGES:
        url = urljoin(BASE_PUBLIC_URL, pkg.name, pkg.directory, pkg.fileName)
    else:
        url = urljoin(config.baseURL, pkg.name, pkg.directory, pkg.fileName)

    # check if we have already downloaded the selected package
    fileName = os.path.join(config.downloadDir, pkg.fileName)
    pkgName, _ = utils.getPackageName(pkg)
    try:
        utils.download(url, fileName, pkgName, config, pkg.dscFile)

    except Exception as err:
        log.logError(
            "The requested file can not be downloaded. "
            f"{urljoin(url, pkg.fileName)} {err}"
        )
        return False

    return True


def downloadLicense(pkg, config):
    if not utils.createDirectory(config.downloadLicenseDir):
        return False

    if not os.access(config.downloadLicenseDir, os.W_OK):
        log.logError(f"Permission denied on directory {config.downloadLicenseDir}")
        sys.exit(0)

    url = urljoin(config.baseURL, "licenses")

    licenseFileName = getLicenseFileName(pkg)

    # download mandatory certificate file
    fn = licenseFileName + CRT_EXTENSION
    try:
        if not utils.download(
            urljoin(url, fn), os.path.join(config.downloadLicenseDir, fn), fn, config
        ):
            log.logError(
                f"Missing {licenseFileName} license. Please get in touch with"
                " info@gempa.de"
            )
            return False

        # download optional license text
        fn = licenseFileName + LICENSE_TXT_EXTENSION
        utils.download(
            urljoin(url, fn), os.path.join(config.downloadLicenseDir, fn), fn, config
        )
    except Exception as err:
        log.logError(
            f"The requested file can not be downloaded. {urljoin(url, fn)} {err}"
        )
        return False

    return True


def getGsmInfo(config):
    response = utils.getDataFromUrl(GSM_INFO_URL, config)
    if not response:
        return

    parsing = False
    for line in response.splitlines():
        if parsing:
            print(line)
        if line.startswith('[" + version + "]'):
            parsing = True


def getLatestRelease(config):
    latestRelease = None
    try:
        releases = os.listdir(os.path.join(config.syncDir, SC))
        releases.remove(NIGHTLY)

        latestRelease = sorted(releases)[len(releases) - 1]

    except Exception:
        pass

    return latestRelease


def isNewReleaseAvailable(config):
    ret = False
    info = None

    if config.release == NIGHTLY:
        return (False, info)

    latestRelease = getLatestRelease(config)
    if not latestRelease:
        return (False, info)

    try:
        if config.isSC3():
            ret = True
        else:
            if int(latestRelease) > int(config.release):
                ret = True
        if ret:
            info = (
                f"New SeisComP release {latestRelease} is available. "
                "Run ./gsm do-release-upgrade to upgrade."
            )
    except Exception:
        pass

    return (ret, info)


def updateCheck(localCache, syncCache, config):
    retCode = 0
    localPackageInfo = localCache.getPackage("gsm")
    if not localPackageInfo:
        return (False, retCode)
    if not config.offline:
        getGsmInfo(config)

    try:
        remotePackageInfo = syncCache.getPackages("gsm")[0]
        if not remotePackageInfo:
            return (False, retCode)

        # Exclude dev version from updates
        if localPackageInfo.version == "VERSION":
            return (False, retCode)

        if localPackageInfo.version < remotePackageInfo.version:
            print(
                f"\nNew gsm release '{remotePackageInfo.version}' available. To upgrade"
                " run: ./gsm install gsm\n"
            )
            if not config.assumeYes:
                ret = checkInput("Do you want to update gsm first?", "y", "yn")
                if ret:
                    retCode = 10
                return (ret, retCode)

            return (False, retCode)
    except IndexError:
        pass

    return (False, retCode)


def fetchChangeLog(pkg, config):
    relPath = pkg.name
    if pkg.filePath:
        relPath = utils.urljoin(os.path.dirname(pkg.filePath), pkg.name)
    if pkg.name in PUBLIC_PACKAGES:
        url = urljoin(BASE_PUBLIC_URL, relPath, CHANGELOG)
    else:
        url = urljoin(config.baseURL, relPath, CHANGELOG)
    response = utils.getDataFromUrl(url, config)

    return response


def fetchChangeLogFromTar(pkg, config):
    fileName = os.path.join(config.downloadDir, pkg.fileName)
    prefix = config.InstallPrefix(config.release)
    try:
        with tarfile.open(fileName, "r", encoding=FILE_ENCODING) as pkgTar:
            file = pkgTar.extractfile(os.path.join(prefix, pkg.changeLog))
            response = ""
            for line in file.readlines():
                response += line.decode(FILE_ENCODING)
    except Exception:
        return None

    return response


def showChangeLog(packages, config, localCache, syncCache, fromFile, version=False):
    responses = []
    if not version:
        if len(packages) == 1:
            pkg = localCache.getPackage(packages[0])
            if not pkg:
                print(f"Package not installed: {packages[0]}")
                return True, False
            if fromFile:
                response = fetchChangeLogFromTar(pkg, config)
            else:
                response = fetchChangeLog(pkg, config)
            if response:
                utils.displayTextSideBySide(response)
                return True, False

    else:
        isShowChangeLog = False
        packages = utils.sorting(packages)
        progress = tqdm(
            packages,
            ascii=True,
            bar_format=utils.getTerminalSizeBarFormat(config),
        )
        progress.set_description(f"{log.getLogPrefix('CHECKING')}")
        for member in progress:
            pkg = localCache.getPackage(member)
            if pkg:
                pkgName = f"{CHANGELOG} {pkg.name}"
            else:
                pkg = syncCache.getPackages(member)[0]
                pkgName = f"{CHANGELOG} {pkg.name}"
            progress.set_description(
                f"{log.getLogPrefix('CHECKING')}{pkgName: <{config.maxlenPackageFullName:}}"
            )
            if fromFile:
                response = fetchChangeLogFromTar(pkg, config)
            else:
                response = fetchChangeLog(pkg, config)
            if response:
                responses.append([response, pkg.version, pkg.name])
            done = f"{CHANGELOG} ({len(progress)}/{len(progress)})"
            progress.set_description(
                f"{log.getLogPrefix(CHECKING_OK)}{done: <{config.maxlenPackageFullName:}}"
            )

        for response in responses:
            try:
                lastUpdate = datetime.strptime(
                    str(response[1]), DEFAULT_UPDATE_DATE_FORMAT
                )
                date = None
                first = True
                showLine = False
                for line in response[0].splitlines():
                    if line.startswith("## "):
                        try:
                            tok = line.split()[1].strip()
                            try:
                                date = datetime.strptime(
                                    tok, DEFAULT_UPDATE_DATE_FORMAT
                                )
                            except Exception:
                                date = datetime.strptime(tok, DEFAULT_TIME_FORMAT)
                        except Exception:
                            pass

                    if not date or date < lastUpdate:
                        continue
                    if line.startswith("### Important"):
                        isShowChangeLog = True
                        if first:
                            print(f"\nImportant changes for {response[2]}\n")
                            first = False
                            showLine = True
                    elif line.startswith("#"):
                        showLine = False

                    if showLine:
                        print(line)

            except Exception:
                pass

    return True, isShowChangeLog


def getLatestStableReleases(config):
    try:
        releasesSC = os.listdir(os.path.join(config.syncDir, SC))
        releasesFinal = []
    except OSError:
        releasesSC = []
        releasesFinal = []
        return False

    try:
        releasesFinal = []
        for release in sorted(releasesSC, reverse=True):
            if release >= config.release or release == NIGHTLY:
                continue

            releasesFinal.append(release)

    except Exception:
        releasesFinal.extend(releasesSC)

    return releasesFinal


def getLatestReleases(config):
    try:
        releasesSC3 = os.listdir(os.path.join(config.syncDir, SC3))
    except OSError:
        releasesSC3 = []

    try:
        releasesSC = os.listdir(os.path.join(config.syncDir, SC))
    except OSError:
        releasesSC = []
        return False

    releasesFinal = []

    if config.release == NIGHTLY:
        return releasesFinal

    for release in sorted(releasesSC3, reverse=True):
        if release == config.release:
            break

        releasesFinal.append(release)

    try:
        if int(config.release) >= 4:
            releasesFinal = []
            for release in sorted(releasesSC, reverse=True):
                if release == config.release:
                    break

                releasesFinal.append(release)

    except Exception:
        releasesFinal.extend(releasesSC)

    return releasesFinal


def showReleases(config):
    try:
        releasesSC3 = os.listdir(os.path.join(config.syncDir, SC3))
    except OSError:
        releasesSC3 = []

    try:
        releasesSC = os.listdir(os.path.join(config.syncDir, SC))
    except OSError:
        return False

    if releasesSC3:
        print("SeisComP3 releases:")
        for release in sorted(releasesSC3, reverse=True):
            print(f"  {release}{'*' if release == config.release else ''}")

    if releasesSC:
        if releasesSC3:
            print("")
        print("SeisComP releases:")
        for release in sorted(releasesSC, reverse=True):
            print(f"  {release}{'*' if release == config.release else ''}")

    return True
