############################################################################
# Copyright (C) 2019 by gempa GmbH                                         #
#                                                                          #
# All Rights Reserved.                                                     #
#                                                                          #
# NOTICE: All information contained herein is, and remains                 #
# the property of gempa GmbH and its suppliers, if any. The intellectual   #
# and technical concepts contained herein are proprietary to gempa GmbH    #
# and its suppliers.                                                       #
# Dissemination of this information or reproduction of this material       #
# is strictly forbidden unless prior written permission is obtained        #
# from gempa GmbH.                                                         #
#                                                                          #
#  Author: Enrico Ellguth, Ralph Henneberger                               #
#  Email: enrico.ellguth@gempa.de, ralix@gempa.de                          #
############################################################################

import os

from tqdm import tqdm

from libs import package, log, utils
from libs.cache import synccache
from libs.const import PT, ANY_NOARCH_DIR, REMOVE_OK, REMOVE
from libs.utils import checkInput, getInput


class ReleaseDowngrade(object):
    def __init__(self, config, localCache):
        self.config = config
        self.localCache = localCache
        self.syncCache = None
        self.installedPackages = None
        self.newRelease = None

    def findObsoletePackages(self):
        config = self.config
        config.release = self.newRelease
        self.syncCache = synccache.SyncCache(config)
        if not self.syncCache.getIndex():
            return False

        obsoletePackages = []
        for name in self.installedPackages:
            if not self.syncCache.contains([name], False):
                obsoletePackages.append(name)

        if obsoletePackages:
            log.logInfo(
                "The following package(s) are no longer available in the new release"
                f" ({self.newRelease}): {', '.join(obsoletePackages)}"
            )
            if checkInput("Do you want to continue?", "n", "yn"):
                return False
        print("")
        return True

    def prepareDowngrade(self):
        releases = package.getLatestStableReleases(self.config)
        count = 1

        if releases:
            print(
                "Please downgrade releases only if you know what you are doing\n"
                "Depending on the release, additional steps may be required, e.g.:\n"
                "\n"
                "* Lower the database schema version\n"
                "* Install additional modules which functionality is part of the newer"
                " SeisComP release\n"
                "* Install required dependencies\n"
                "\n"
                "If your SeisComP installation is not under version control, we"
                " recommend to make a backup before continuing.\n"
            )
            if checkInput("Do you want to continue?", "n", "yn"):
                return False
        else:
            print("No older release found.")
            return False

        print("Available releases:")
        print("")

        for release in releases:
            print(f"{count}. {release}")
            count += 1
        print("")

        try:
            idx = getInput("Downgrade to Option")
            value = int(idx)
            self.newRelease = releases[value - 1]

        except KeyboardInterrupt:
            return False

        except Exception:
            print("Invalid entry. Please select an entry from the list.")
            return False

        packages = list(self.localCache.getIndex().keys())
        self.installedPackages = []

        for name in packages:
            pkg = self.localCache.getPackage(name)
            if pkg.type == PT.APP:
                self.installedPackages.append(pkg.name)

        if not self.runModuleCheck():
            return False

        if not self.findObsoletePackages():
            return False

        if not self.downloadNewPackages():
            return False

        stopUpgrade = False
        for name in self.installedPackages:
            if not self.localCache.isInstallable(self.localCache.getPackage(name)):
                stopUpgrade = True

        if stopUpgrade:
            log.logError(
                "Could not perform the upgrade because at least one application is not"
                " covered by the license. Please get in touch with info@gempa.de"
            )

            return False

        return True

    def removePackages(self):
        members = utils.sorting(self.installedPackages)
        count = 0
        progress = tqdm(
            members,
            ascii=True,
            bar_format=utils.getTerminalSizeBarFormat(self.config),
        )
        for member in progress:
            pkgName, _ = utils.getPackageName(self.localCache.getPackage(member))
            progress.set_description(
                f"{log.getLogPrefix(REMOVE)}{pkgName: <{self.config.maxlenPackageFullName:}}"
            )
            count = count + 1
            done = f"DONE ({len(progress)}/{count})"
            self.localCache.removePackage(member, False)
            if count == len(members):
                progress.set_description(
                    f"{log.getLogPrefix(REMOVE_OK)}"
                    f"{done: <{self.config.maxlenPackageFullName:}}"
                )

        self.config.update({"release": self.newRelease}, [])
        self.config.updateInstallInfoFile("release", self.newRelease)

        return True

    def installPackages(self, app):
        installPackages = []
        app.init()

        self.config.assumeYes = True
        for name in self.installedPackages:
            if self.syncCache.contains([name], False):
                installPackages.append(name)

        return app.install(installPackages, True)

    def downloadNewPackages(self):
        downloadPackages = []

        self.config.setArchDir(
            os.path.join(
                self.config.release,
                self.config.os,
                self.config.os_version,
                self.config.arch,
            )
        )
        self.config.setNoArchDir(os.path.join(self.config.release, ANY_NOARCH_DIR))

        syncCache = synccache.SyncCache(self.config)

        if not self.installedPackages:
            print("\nNo installed packages found.")
        for name in self.installedPackages:

            if syncCache.contains([name], False):
                downloadPackages.append(name)

        utils.setMaxlenPackageFullName(
            downloadPackages, self.localCache, self.syncCache, self.config
        )

        if not self.config.offline:
            utils.checkUrlAccessProgress(self.config, syncCache, downloadPackages)

            ret = syncCache.downloadPackages(downloadPackages)
        else:
            ret = True
        failed = []
        fileNotExist = []
        if ret:
            for name in downloadPackages:
                pkgSync = syncCache.getPackages(name)[0]
                fileName = os.path.join(self.config.downloadDir, pkgSync.fileName)
                if os.path.exists(fileName):
                    fail = utils.hashfile(
                        self.config, fileName, pkgSync.checksum, pkgSync
                    )
                    if not fail:
                        failed.append(pkgSync.name)
                        if os.path.isfile(fileName):
                            os.remove(fileName)
                else:
                    fileNotExist.append(pkgSync)

        if fileNotExist:
            for pkg in fileNotExist:
                print(
                    f"The {pkg.fileName} file does not exist in the"
                    f" {self.config.downloadDir} directory."
                )
                return False

        if failed:
            print(
                "\nChecksum validation for the following packages failed:"
                f" {', '.join(failed)}"
            )
            print("\nTry to run ./gsm update to refresh the package index.")

            return False

        return ret

    def runModuleCheck(self):
        syncCache = synccache.SyncCache(self.config)

        if not syncCache.getIndex():
            return False

        for name in self.installedPackages:
            try:
                syncCache.getPackages(name)[0]
            except Exception:
                log.logError(
                    f"Can't run release downgrade! module {name} is not available for"
                    f" release {self.newRelease}"
                )
                return False

        return True

    def run(self, app):
        self.config.updateOrDowngrade = True
        if not self.prepareDowngrade():
            return False
        if not self.config.downloadOnly:
            if not self.removePackages():
                return False

            return self.installPackages(app)

        return True
