############################################################################
# Copyright (C) 2019 by gempa GmbH                                         #
#                                                                          #
# All Rights Reserved.                                                     #
#                                                                          #
# NOTICE: All information contained herein is, and remains                 #
# the property of gempa GmbH and its suppliers, if any. The intellectual   #
# and technical concepts contained herein are proprietary to gempa GmbH    #
# and its suppliers.                                                       #
# Dissemination of this information or reproduction of this material       #
# is strictly forbidden unless prior written permission is obtained        #
# from gempa GmbH.                                                         #
#                                                                          #
#  Author: Enrico Ellguth, Ralph Henneberger                               #
#  Email: enrico.ellguth@gempa.de, ralix@gempa.de                          #
############################################################################

import os
import shutil

from tqdm import tqdm

from libs import package, log, utils
from libs.cache import synccache
from libs.const import (
    BASE_PUBLIC_URL,
    NIGHTLY,
    SC,
    SC3,
    PT,
    ANY_NOARCH_DIR,
    REMOVE,
    REMOVE_OK,
    RENAME_SEEDLINK_BUFFER,
    CHECK_SCCONFIG,
    SEISCOMP_TOOLS,
)
from libs.utils import checkInput, getInput


class ReleaseUpgrade(object):
    def __init__(self, config, localCache):
        self.config = config
        self.localCache = localCache
        self.syncCache = None
        self.oldPath = None
        self.newPath = None
        self.oldUserConfigPath = None
        self.newUserConfigPath = None
        self.installedPackages = None
        self.newRelease = None
        self.moveRelease = False
        self.oldRelease = config.release

    def createInstallPath(self):
        self.oldPath = self.config.installPath
        newPrefix = self.config.InstallPrefix(self.newRelease)
        self.newPath = os.path.join(self.config.installDir, newPrefix)

        print("")
        self.newPath = getInput("New install directory", self.newPath)
        if os.path.exists(self.newPath):
            print("")
            log.logError("A file or directory with the install path already exists")
            return False

        self.oldUserConfigPath = getInput(
            "Current user config directory", self.config.userConfigDir
        )
        self.newUserConfigPath = getInput(
            "New user config directory",
            os.path.join(self.config.homeDir, f".{newPrefix}"),
        )

        try:
            os.makedirs(self.newPath)
            os.rmdir(self.newPath)
        except OSError:
            print("")
            log.logError(f"Can't create new install directory {self.newPath} ...")
            return False

        return True

    def findObsoletePackages(self):
        config = self.config
        config.release = self.newRelease
        self.syncCache = synccache.SyncCache(config)
        if not self.syncCache.getIndex():
            return False

        obsoletePackages = []
        for name in self.installedPackages:
            if not self.syncCache.contains([name], False):
                obsoletePackages.append(name)

        if self.moveRelease:
            try:
                obsoletePackages.remove(SC3)
            except ValueError:
                pass

        if obsoletePackages:
            log.logInfo(
                "The following package(s) are no longer available in the new release"
                f" ({self.newRelease}): {', '.join(obsoletePackages)}"
            )
            if checkInput("Do you want to continue?", "n", "yn"):
                return False
        print("")
        return True

    def prepareUpgrade(self):
        releases = package.getLatestReleases(self.config)
        count = 1

        if releases:
            print(
                "\nIf your SeisComP installation is not under version control, we"
                " recommend to make a backup before continuing."
            )
            if checkInput("Do you want to continue?", "n", "yn"):
                return False
        else:
            print("No new release found.")
            return False

        print("Available releases:")
        print("")

        for release in releases:
            print(f"{count}. {release}")
            count += 1
        print("")

        try:
            idx = getInput("Upgrade to Option")
            value = int(idx)
            self.newRelease = releases[value - 1]

        except KeyboardInterrupt:
            return False

        except Exception:
            print("Invalid entry. Please select an entry from the list.")
            return False

        packages = list(self.localCache.getIndex().keys())
        self.installedPackages = []

        try:
            if self.config.isSC3() and (
                self.newRelease == NIGHTLY or int(self.newRelease) >= 4
            ):
                self.moveRelease = True
        except ValueError:
            pass

        for name in packages:
            pkg = self.localCache.getPackage(name)
            if pkg.type == PT.APP:
                self.installedPackages.append(pkg.name)

        if not self.runModuleCheck():
            return False

        if not self.findObsoletePackages():
            return False

        if self.moveRelease:
            if not self.createInstallPath():
                return False

        if not self.downloadNewPackages():
            return False

        stopUpgrade = False
        for name in self.installedPackages:
            if not self.localCache.isInstallable(self.localCache.getPackage(name)):
                stopUpgrade = True

        if stopUpgrade:
            log.logError(
                "Could not perform the upgrade because at least one application is not"
                " covered by the license. Please get in touch with info@gempa.de"
            )

            return False

        return True

    def removePackages(self):
        members = utils.sorting(self.installedPackages)
        count = 0
        progress = tqdm(
            members,
            ascii=True,
            bar_format=utils.getTerminalSizeBarFormat(self.config),
        )
        for member in progress:
            pkgName, _ = utils.getPackageName(self.localCache.getPackage(member))
            progress.set_description(
                f"{log.getLogPrefix(REMOVE)}{pkgName: <{self.config.maxlenPackageFullName:}}"
            )
            count = count + 1
            done = f"DONE ({len(progress)}/{count})"
            self.localCache.removePackage(member, False)
            if count == len(members):
                progress.set_description(
                    f"{log.getLogPrefix(REMOVE_OK)}"
                    f"{done: <{self.config.maxlenPackageFullName:}}"
                )

        self.config.update({"release": self.newRelease}, [])
        self.config.updateInstallInfoFile("release", self.newRelease)

        if self.moveRelease:
            try:
                if os.path.isdir(self.oldPath):
                    shutil.move(self.oldPath, self.newPath)
            except Exception:
                log.logError(f"Can't move {self.oldPath} to {self.newPath}")
                return False

            try:
                if os.path.isdir(self.oldUserConfigPath):
                    shutil.move(self.oldUserConfigPath, self.newUserConfigPath)
            except Exception:
                log.logError(
                    f"Can't move {self.oldUserConfigPath} to {self.newUserConfigPath}"
                )
                return False

            self.config.update({"install_path": self.newPath}, [])

        return True

    def installPackages(self, app):
        installPackages = []
        app.init()
        try:
            if int(self.newRelease) >= 4 and SC3 in self.installedPackages:
                installPackages.append(SC)
        except Exception:
            pass

        if self.newRelease == NIGHTLY and SC3 in self.installedPackages:
            installPackages.append(SC)

            print("")
            print("Please check your SeisComP environment settings in ~/.bashrc")
            print("")
            print(
                "SeedLink in SeisComP3 prior to 2020 uses a simplified buffer. Migrate"
                f" the buffer using '{RENAME_SEEDLINK_BUFFER}' after upgrading."
            )
            print(
                "Download the script here:"
                f" {BASE_PUBLIC_URL}/tools/{RENAME_SEEDLINK_BUFFER}"
            )

        print("")
        print(
            f"Please run the script '{RENAME_SEEDLINK_BUFFER}' after upgrading to check"
            " the SeisComP configuration for deprecated parameters."
        )
        print(
            "The tool can be download here:"
            f" {BASE_PUBLIC_URL}/tools/{CHECK_SCCONFIG}."
        )
        print("")
        print("Please also check your SeisComP environment settings in ~/.bashrc")
        print("")

        self.config.assumeYes = True
        for name in self.installedPackages:
            if self.syncCache.contains([name], False):
                installPackages.append(name)

        return app.install(installPackages, True)

    def downloadNewPackages(self):
        downloadPackages = []

        self.config.setArchDir(
            os.path.join(
                self.config.release,
                self.config.os,
                self.config.os_version,
                self.config.arch,
            )
        )
        self.config.setNoArchDir(os.path.join(self.config.release, ANY_NOARCH_DIR))

        syncCache = synccache.SyncCache(self.config)

        try:
            if int(self.newRelease) >= 4 and SC3 in self.installedPackages:
                downloadPackages.append(SC)
        except Exception:
            pass

        if self.newRelease == NIGHTLY and SC3 in self.installedPackages:
            downloadPackages.append(SC)

        if not self.installedPackages:
            print("\nNo installed packages found.")
        for name in self.installedPackages:

            if syncCache.contains([name], False):
                downloadPackages.append(name)

        utils.setMaxlenPackageFullName(
            downloadPackages, self.localCache, self.syncCache, self.config
        )
        if not self.config.offline:
            utils.checkUrlAccessProgress(self.config, syncCache, downloadPackages)

            ret = syncCache.downloadPackages(downloadPackages)
        else:
            ret = True

        failed = []
        fileNotExist = []
        if ret:
            for name in downloadPackages:
                pkgSync = syncCache.getPackages(name)[0]
                fileName = os.path.join(self.config.downloadDir, pkgSync.fileName)
                if os.path.exists(fileName):
                    fail = utils.hashfile(
                        self.config, fileName, pkgSync.checksum, pkgSync
                    )
                    if not fail:
                        failed.append(pkgSync.name)
                        if os.path.isfile(fileName):
                            os.remove(fileName)
                else:
                    fileNotExist.append(pkgSync)

        if fileNotExist:
            for pkg in fileNotExist:
                print(
                    f"The {pkg.fileName} file does not exist in the"
                    f" {self.config.downloadDir} directory."
                )
                return False
        if failed:
            print(
                "\nChecksum validation for the following packages failed:"
                f" {', '.join(failed)}"
            )
            print("\nTry to run ./gsm update to refresh the package index.")

            return False

        return ret

    def runModuleCheck(self):
        syncCache = synccache.SyncCache(self.config)

        if not syncCache.getIndex():
            return False

        for name in self.installedPackages:
            try:
                syncCache.getPackages(name)[0]
            except Exception:
                log.logError(
                    f"Can't run release upgrade! module {name} is not available for"
                    f" release {self.config.release}"
                )
                return False

        return True

    def installSeiscompTools(self):
        syncCache = synccache.SyncCache(self.config)
        pkg = syncCache.getPackages(SEISCOMP_TOOLS)[0]
        url = utils.getDownloadDirectoryURL(pkg.name, pkg, self.config)
        ret, _ = utils.checkUrlAccess(url, self.config.user, self.config.password)

        if not self.config.offline and ret:
            if not self.localCache.isInstalled(pkg, False):
                print("\n")
                if checkInput(
                    "Do you want install seiscomp-tools"
                    f" ({RENAME_SEEDLINK_BUFFER} and"
                    f" {CHECK_SCCONFIG}) ?",
                    "Y",
                    "yn",
                ):
                    download = syncCache.downloadPackages([pkg.name])
                    if not download:
                        print("\nThe download of all packages failed.")
                    if not syncCache.checksumCheck(pkg.name):
                        print(
                            "\nTry to run ./gsm update to refresh the package index"
                            f" and run ./gsm install {SEISCOMP_TOOLS}, after the upgrade..."
                        )
                    else:
                        ret, _ = self.localCache.installPackage(pkg.name, None)

    def run(self, app):
        self.config.updateOrDowngrade = True
        if not self.prepareUpgrade():
            return False
        if not self.config.downloadOnly:
            if not self.removePackages():
                return False

            ret = self.installPackages(app)
            self.installSeiscompTools()
            return ret
        return True
