############################################################################
# Copyright (C) 2019 by gempa GmbH                                         #
#                                                                          #
# All Rights Reserved.                                                     #
#                                                                          #
# NOTICE: All information contained herein is, and remains                 #
# the property of gempa GmbH and its suppliers, if any. The intellectual   #
# and technical concepts contained herein are proprietary to gempa GmbH    #
# and its suppliers.                                                       #
# Dissemination of this information or reproduction of this material       #
# is strictly forbidden unless prior written permission is obtained        #
# from gempa GmbH.                                                         #
#                                                                          #
#  Author: Enrico Ellguth, Ralph Henneberger                               #
#  Email: enrico.ellguth@gempa.de, ralix@gempa.de                          #
############################################################################

import hashlib
import os
import platform
import subprocess
import shutil
import sys
import tarfile
import time
import math

from datetime import datetime, date
from subprocess import Popen, PIPE
from zipfile import ZipFile

import humanize
import pytz
import requests

import numpy

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from requests.auth import HTTPBasicAuth
from tqdm import tqdm
from tzlocal import get_localzone

from libs.const import (
    LICENSES,
    OS_MAPPING,
    SC,
    SC3,
    urljoin,
    DEFAULT_TIMEOUT,
    BASE_PUBLIC_URL,
    PT,
    CUSTOM_INSTALL_PATH,
    DEFAULT_TIME_FORMAT,
    DEFAULT_UPDATE_DATE_FORMAT,
    FILE_ENCODING,
    MAX_LEN_PACKAGE_FULL_NAME,
    MBTILES_EXTENSION,
    GRD_EXTENSION,
    ZIP_EXTENSION,
    TARGZ_EXTENSION,
    TAR_EXTENSION,
    DB_EXTENSION,
    CRT_EXTENSION,
    LICENSE_TXT_EXTENSION,
    CHANGELOG,
    BAR_RIGHT_LENGHT,
    INSTALL,
    INSTALL_OK,
    INSTALL_FAILED,
    GET,
    GET_OK,
    CHECKSUM,
    CHECKSUM_OK,
    CHECKSUM_FAILED,
    CHECKING_OK,
    CHECKING_FAILED,
    ACCESS,
    DEFAULT_ARCH,
    PUBLIC_PACKAGES,
)

from libs import log
from libs.cache import custominstall

try:
    real_raw_input = raw_input
except NameError:
    real_raw_input = input


# request and optionally enforce user input
# @param question The question to be answered.
# @param default The default value to use if no input was made
# @param options List or string of available options. If defined the input must
#        match one of the options (unless a default is specified). If the input
#        is invalid the question is repeated.
def getInput(question, default=None, options=None):
    def _default(text):
        # print default value to previous line if no input was made
        if default:
            print(f"\x1b[F\x1b[{len(text) + 1}G{default}")
        return default

    # no options: accept any type of input
    if not options:
        defaultStr = "" if default is None else f" [{default}]"
        question = f"{question}{defaultStr}: "
        return real_raw_input(question) or _default(question)

    if default is not None:
        default = str(default).lower()

    # options supplied: check and enforce input
    opts = [str(o).lower() for o in options]
    optStr = "/".join(o.upper() if o == default else o for o in opts)
    question = f"{question} [{optStr}]: "
    while True:
        res = real_raw_input(question)
        if not res and default:
            return _default(question)

        if res.lower() in opts:
            return res.lower()


def sorting(lst):
    lenlist = [len(x) for x in lst]
    sortedindex = numpy.argsort(lenlist)
    tempLst = [lst[i] for i in sortedindex]
    return tempLst


def checkUrlAccessProgress(config, syncCache, members):
    members = sorting(members)
    failed = []
    progress = tqdm(
        members,
        ascii=True,
        bar_format=getTerminalSizeBarFormat(config),
    )
    progress.set_description(f"{log.getLogPrefix(ACCESS)}")
    for member in progress:
        pkg = syncCache.getPackages(member)[0]
        pkgName, _ = getPackageName(pkg)
        done = f"{ACCESS} ({len(members)}/{len(members) - len(failed)})"
        progress.set_description(
            f"{log.getLogPrefix(ACCESS)}{pkgName: <{config.maxlenPackageFullName:}}"
        )
        url = getDownloadDirectoryURL(pkg.name, pkg, config)
        ret, _ = checkUrlAccess(url, config.user, config.password)

        if not ret:
            failed.append(pkg.name)
            progress.set_description(
                f"{log.getLogPrefix(CHECKING_FAILED)}{done: <{config.maxlenPackageFullName:}}"
            )
        else:
            progress.set_description(
                f"{log.getLogPrefix(CHECKING_OK)}{done: <{config.maxlenPackageFullName:}}"
            )

    return failed


def checkUrlAccess(baseURL, user, password):
    try:
        r = requests.get(
            baseURL,
            stream=True,
            auth=HTTPBasicAuth(user, password),
            timeout=DEFAULT_TIMEOUT,
        )
        if r.status_code == 200:
            return (True, r.status_code)
    except Exception:
        return (True, 200)

    return (False, r.status_code)


def isReleaseAvailable(release):
    try:
        endpoints = [
            SC3,
            SC,
        ]
        for endpoint in endpoints:
            url = urljoin(BASE_PUBLIC_URL, endpoint, release, "/")
            response = requests.head(url, timeout=DEFAULT_TIMEOUT, allow_redirects=True)
            if response.status_code == 200:
                return True
    except Exception:
        return True

    return False


def checkInput(question, default, options):
    default = str(default).lower()
    return getInput(question, default, options) == default


def humanbytes(B):
    try:
        return humanize.naturalsize(B, binary=True)
    except ValueError:
        return "N/A"


def is64bit():
    return platform.machine().endswith("64")


def checkOS(config):
    osName, version, arch = detectOS()
    if osName != config.os or version != config.os_version or arch != config.arch:
        print(
            f"{log.getLogPrefix('NOTICE')}Selected operating system"
            " is not compatible with this system."
        )
        print(
            f"{log.getLogPrefix('NOTICE')}Please consider to change"
            " the operating system with './gsm switch-os'.\n"
        )

    return True


def detectOS():
    try:
        arch = platform.machine()
    except BaseException:
        arch = DEFAULT_ARCH

    try:
        os_release = {}
        with open("/etc/os-release", "r", encoding=FILE_ENCODING) as f:
            for line in f:
                toks = line.split("=")
                if len(toks) != 2:
                    continue
                os_release[toks[0].strip().upper()] = toks[1].strip()

        osName = OS_MAPPING.get(os_release["ID"].strip('"'))
        if not osName:
            osName = os_release["ID"].strip('"')

        version = os_release["VERSION_ID"].strip('"')
        if osName == "rhel":
            try:
                version = str(math.floor(float(version)))
            except Exception:
                pass

        return osName, version, arch

    except BaseException:
        return "rhel", "7", arch


def du(path):
    if os.path.exists(path):
        return (
            subprocess.check_output(["du", "-sh", path])
            .split()[0]
            .decode(FILE_ENCODING)
        )
    return None


def toUTC(dt):
    timeZone = get_localzone()
    local = timeZone.localize(dt)
    return local.astimezone(pytz.utc)


def readCertificate(name, config):
    fileName = f"{name}{CRT_EXTENSION}"
    licenseFiles = [
        os.path.join(config.downloadLicenseDir, fileName),
        os.path.join(config.licenseDir, fileName),
        os.path.join(config.userConfigDir, LICENSES, fileName),
    ]

    cert = None

    for licenseFile in licenseFiles:
        if os.path.isfile(licenseFile):
            res = readCertificateFile(licenseFile)
            if res and (
                cert is None
                or (
                    res.updatesUntil
                    and (not cert.updatesUntil or cert.updatesUntil < res.updatesUntil)
                )
            ):
                cert = res
                break

    return cert


class CertLocal(object):
    def __init__(self, issuer, not_valid_after):
        self.issuer = issuer
        self.not_valid_after = not_valid_after
        self.updatesUntil = None


def readCertificateFile(licenseFile):
    try:
        with open(licenseFile, "rb") as certFile:
            pemDataFile = certFile.read()
            cert = x509.load_pem_x509_certificate(pemDataFile, default_backend())
            if not hasattr(cert, "updatesUntil"):
                # setattr(cert, "updatesUntil", None)
                if hasattr(cert, "not_valid_after_utc"):
                    cert = CertLocal(cert.issuer, cert.not_valid_after_utc)
                else:
                    cert = CertLocal(cert.issuer, cert.not_valid_after)

            for i in str(cert.issuer).split("="):
                if not "updatesUntil" in i:
                    continue
                timeStr = str(cert.issuer).split("updatesUntil=")[1].split(" ")[0]
                cert.updatesUntil = datetime.strptime(
                    timeStr, DEFAULT_TIME_FORMAT
                ).strftime(DEFAULT_UPDATE_DATE_FORMAT)
    except Exception:
        return None

    return cert


def getTerminalLines():
    lines = shutil.get_terminal_size().lines
    return lines


def getTerminalColumns():
    columns = shutil.get_terminal_size().columns
    return columns


def displayTextSideBySide(text):
    maxLines = getTerminalLines()

    lines = text.split("\n")

    for i in range(0, len(lines), maxLines):
        chunk = "\n".join(lines[i : i + maxLines])
        print(chunk)
        if i + maxLines < len(lines):
            input("\nPress any button to continue or (Ctrl+C) to cancel...")


def getTerminalSizeBarFormat(config):
    barSize = getTerminalColumns() - config.maxlenPackageFullName - BAR_RIGHT_LENGHT
    barFormat = "{l_bar}{bar:" + f"{barSize}" + "}{r_bar}"
    return barFormat


def hashfile(config, fileName, checksum, pkg):
    pkgName, _ = getPackageName(pkg)
    blockSize = 2**16  # 64 KiB
    hasher = hashlib.sha256()
    if not os.path.isfile(fileName):
        log.logError(f"No such file or directory: {fileName}")
        return False
    fileSize = os.path.getsize(fileName)
    try:
        with open(fileName, "rb") as f:
            with tqdm(
                total=fileSize,
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
                desc=f"{log.getLogPrefix(CHECKSUM)}{pkgName: <{config.maxlenPackageFullName:}}",
                ascii=True,
                bar_format=getTerminalSizeBarFormat(config),
            ) as pbar:
                for buf in iter(lambda: f.read(blockSize), b""):
                    hasher.update(buf)
                    pbar.update(len(buf))

                if hasher.hexdigest() != checksum:
                    pbar.total = 1
                    pbar.n = 1
                    pbar.refresh()
                    pbar.set_description(
                        f"{log.getLogPrefix(CHECKSUM_FAILED)}"
                        f"{pkgName: <{config.maxlenPackageFullName:}}"
                    )
                    return False

                pbar.set_description(
                    f"{log.getLogPrefix(CHECKSUM_OK)}{pkgName: <{config.maxlenPackageFullName:}}"
                )
    except IOError as e:
        print(f"Error reading file: {e}")
        return False

    return True


def isDirty(fileName, url, config):
    lastModified = None
    try:
        response = requests.head(
            url,
            auth=(config.user, config.password),
            timeout=DEFAULT_TIMEOUT,
            allow_redirects=True,
        )
        if response.status_code != 200:
            raise Exception(f"{url}: {response.status_code} {response.reason}")
        try:
            lastModifiedStr = response.headers["last-modified"]
            lastModified = datetime.strptime(
                lastModifiedStr, "%a, %d %b %Y %H:%M:%S %Z"
            )
            lastModified.replace(tzinfo=pytz.UTC)
            last_update = datetime.fromtimestamp(
                os.path.getmtime(os.path.abspath(fileName))
            )
            last_update.replace(tzinfo=pytz.UTC)
            if last_update >= lastModified:
                return False, lastModified
        except KeyError:
            pass
    except OSError:
        pass

    return True, lastModified


def getDataFromUrl(url, config, update=False):
    try:
        r = requests.get(
            url, auth=(config.user, config.password), timeout=DEFAULT_TIMEOUT
        )
        r.raise_for_status()
        return r.text
    except requests.exceptions.HTTPError as err:
        if update is False:
            log.logError(f"{url}: ({err})")
    except requests.exceptions.RequestException as err:
        log.logError(f"{err}")

    return None


def download(url, fileName, desc, config, dscFile=None):
    if config.showUrls:
        desc = url
    _, retCode = checkUrlAccess(config.baseURL, config.user, config.password)
    if retCode == 401:
        log.logError("Invalid login credentials. Please check your gsm configuration.")
        sys.exit(0)
    if retCode == 404:
        log.logError("Invalid repository URL. Please check your gsm configuration.")
        sys.exit(0)
    if config.forceDownload:
        dirty, mtime = isDirty(fileName, url, config)
    else:
        try:
            dirty, mtime = isDirty(fileName, url, config)
            if not dirty:
                return True
            if not mtime:
                log.logError(
                    "http request did not provide 'last-modified'. Please try again."
                )
                return False
        except Exception as err:
            log.logError(f"{err}")
            return False
    try:
        r = requests.get(
            url,
            stream=True,
            auth=HTTPBasicAuth(config.user, config.password),
            timeout=DEFAULT_TIMEOUT,
        )
    except requests.exceptions.RequestException as err:
        log.logError(f"{str(err)}")
        return False

    fileSize = int(r.headers.get("content-length", 0))

    tmpFile = os.path.join(fileName + "." + "tmp")
    if os.path.exists(tmpFile):
        firstByte = os.path.getsize(tmpFile)
        filemode = "ab"
    else:
        filemode = "wb"
        firstByte = 0
        if firstByte >= fileSize:
            return fileSize

    header = {"Accept-Encoding": "", "Range": f"bytes={firstByte}-{fileSize}"}
    try:
        r = requests.get(
            url,
            headers=header,
            stream=True,
            auth=HTTPBasicAuth(config.user, config.password),
            timeout=DEFAULT_TIMEOUT,
        )
        if r.status_code >= 300:
            log.logError(f"{url}: {r.status_code} {r.reason}")
            return False
    except requests.exceptions.RequestException as err:
        log.logError(f"{str(err)}")
        return False

    with open(tmpFile, filemode) as f:
        with tqdm(
            total=fileSize,
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
            desc=f"{log.getLogPrefix(GET)}{desc: <{config.maxlenPackageFullName:}}",
            initial=firstByte,
            ascii=True,
            bar_format=getTerminalSizeBarFormat(config),
        ) as pbar:
            for chunk in r.iter_content(32 * 1024):
                f.write(chunk)
                pbar.update(len(chunk))
            pbar.set_description(
                f"{log.getLogPrefix(GET_OK)}{desc: <{config.maxlenPackageFullName:}}"
            )
    try:
        if fileSize != 0 and os.path.getsize(tmpFile) != fileSize:
            raise Exception("incomplete download")
    except OSError as err:
        log.logError(f"{err}: {desc}, try to run install again")
        return False

    os.replace(tmpFile, fileName)
    os.utime(fileName, (time.time(), int(mtime.strftime("%s"))))

    if dscFile:
        if os.path.isfile(dscFile):
            shutil.copy(dscFile, config.downloadDir)

    return True


def copy(config, fileName, path, pkgName):
    if not os.path.isfile(fileName):
        log.logError(f"No such file or directory: {fileName}")
        return False
    fileSize = os.path.getsize(fileName)
    bufSize = 2**20 * 10  # 10MiB
    numChunks = math.ceil(fileSize / bufSize)

    try:
        with open(fileName, "rb") as src_f, open((path), "wb") as dest_f:
            progress = tqdm(
                range(numChunks),
                ascii=True,
                bar_format=getTerminalSizeBarFormat(config),
            )
            progress.set_description(
                f"{log.getLogPrefix(INSTALL_OK)}{pkgName: <{config.maxlenPackageFullName:}}"
            )

            for _ in progress:
                chunk = src_f.read(bufSize)
                dest_f.write(chunk)
        progress.set_description(
            f"{log.getLogPrefix(INSTALL_OK)}{pkgName: <{config.maxlenPackageFullName:}}"
        )
    except IOError as err:
        progress.set_description(f"{log.getLogPrefix(INSTALL_FAILED)}{pkgName} ({err})")
        log.logError(f"{pkgName} ({err})")
        return False

    return True


def unzip(config, fileName, path, pkg):
    pkgName, installScript = getPackageName(pkg)

    fileCount = 0
    if not os.path.isfile(fileName):
        log.logError(f"No such file or directory: {fileName}")
        return fileCount
    try:
        with ZipFile(fileName, "r") as zipper:
            members = zipper.infolist()
            progress = tqdm(
                members,
                ascii=True,
                bar_format=getTerminalSizeBarFormat(config),
            )
            progress.set_description(
                f"{log.getLogPrefix(INSTALL)}{pkgName: <{config.maxlenPackageFullName:}}"
            )
            for member in progress:
                # Special treatment for regular files:
                # The tarfile.extract method overrides existing files which
                # causes application crashes in case a binary is still running
                # or a .so file is mapped to a running process. Here we
                # replicate the default behavior or GNU tar which first removes
                # existing files and then extracts the file content to a new
                # file using a different inode and disk space. Running
                # applications will work with the old file version until they
                # are restarted.
                if not member.is_dir():
                    dst = os.path.join(path, member.filename)
                    if not os.path.islink(dst) and os.path.isfile(dst):
                        os.unlink(dst)
                    fileCount += 1
                if installScript:
                    if member.name == installScript:
                        continue
                    zipper.extract(member, path=path)
                else:
                    zipper.extract(member, path=path)
                progress.set_description(
                    f"{log.getLogPrefix(INSTALL_OK)}{pkgName: <{config.maxlenPackageFullName:}}"
                )
    except IOError as err:
        progress.set_description(f"{log.getLogPrefix(INSTALL_FAILED)}{pkgName} ({err})")
        log.logError(f"{pkgName} ({err})")

    return fileCount


def getPackageName(pkg):
    installScript = None
    if hasattr(pkg, "name"):
        pkgName = pkg.name + " " + pkg.version
        installScript = pkg.installScript
    else:
        pkgName = pkg

    return (pkgName, installScript)


def runCustomInstallScript(self, fileName, pkg):
    _, installScript = getPackageName(pkg)

    if not os.path.isfile(fileName):
        log.logError(f"No such file or directory: {fileName}")

        return False, None

    mode = "r"
    if fileName.endswith((TARGZ_EXTENSION, DB_EXTENSION)):
        mode = "r:gz"
    elif fileName.endswith(TAR_EXTENSION):
        mode = "r:"

    # extracts the custom install script from the tarfile to gsm/libs/cache/custominstall
    extractPath = os.path.join(self.config.baseDir, CUSTOM_INSTALL_PATH)
    try:
        with tarfile.open(fileName, mode) as tf:
            tf.extraction_filter = getattr(
                tf, "data_filter", (lambda member, path: member)
            )
            tf.extract(member=installScript, path=extractPath)
    except KeyError:
        return False, "installScript"

    # runs the custom install script and checks if there are files in the
    # file system that should be installed. #
    # It returns the files that exist in the file system
    existFileName = custominstall.checkFilesExist(self.config, pkg)

    # if there are files that exist the package cannot be installed.
    # If this is the case I delete the custom install script from
    # libs/cache/custominstall and the corresponding directory
    # if there are no more custom install scripts there
    if existFileName:
        os.remove(os.path.join(extractPath, installScript))
        shutil.rmtree(os.path.join(extractPath, "__pycache__"))

    if not os.listdir(extractPath):
        shutil.rmtree(extractPath)

    return True, existFileName


def untar(config, fileName, path, pkg):

    installFiles = []

    pkgName, installScript = getPackageName(pkg)

    fileCount = 0
    if not os.path.isfile(fileName):
        log.logError(f"No such file or directory: {fileName}")
        return fileCount
    mode = "r"
    if fileName.endswith((TARGZ_EXTENSION, DB_EXTENSION)):
        mode = "r:gz"
    elif fileName.endswith(TAR_EXTENSION):
        mode = "r:"

    try:
        with tarfile.open(fileName, mode) as tar:
            tar.extraction_filter = getattr(
                tar, "data_filter", (lambda member, path: member)
            )
            members = tar.getmembers()
            progress = tqdm(
                members,
                ascii=True,
                bar_format=getTerminalSizeBarFormat(config),
            )
            progress.set_description(
                f"{log.getLogPrefix(INSTALL)}{pkgName: <{config.maxlenPackageFullName:}}"
            )

            # special case: core.db:
            if fileName.endswith(DB_EXTENSION):
                for member in progress:
                    installFiles.append(member.name)
                    if member.isfile():
                        fileCount += 1
                    tar.extract(member, path=path)
                    if fileCount == len(members):
                        progress.set_description(
                            f"{log.getLogPrefix(INSTALL_OK)}"
                            f"{pkgName: <{config.maxlenPackageFullName:}}"
                        )
                return fileCount, None
            # default
            i = 0
            prefix = config.installPrefix()
            if prefix == config.rootDir:
                prefix = None
            for member in progress:
                installFiles.append(member.name)
                i = i + 1
                # replace path prefix with installation dir
                if prefix:
                    member.name = member.name.replace(prefix, config.rootDir, 1)

                # Special treatment for regular files:
                # The tarfile.extract method overrides existing files which
                # causes application crashes in case a binary is still running
                # or a .so file is mapped to a running process. Here we
                # replicate the default behavior or GNU tar which first removes
                # existing files and then extracts the file content to a new
                # file using a different inode and disk space. Running
                # applications will work with the old file version until they
                # are restarted.
                if member.isfile():
                    dst = os.path.join(path, member.name)
                    if not os.path.islink(dst) and os.path.isfile(dst):
                        os.unlink(dst)
                    fileCount += 1

                if installScript:
                    if member.name == installScript:
                        continue
                    tar.extract(member, path=path)
                else:
                    tar.extract(member, path=path)

                if len(members) == i:
                    progress.set_description(
                        f"{log.getLogPrefix(INSTALL_OK)}"
                        f"{pkgName: <{config.maxlenPackageFullName:}}"
                    )

            return fileCount, installFiles
    except IOError as err:
        progress.set_description(f"{log.getLogPrefix(INSTALL_FAILED)}{pkgName} ({err})")
        log.logError(f"{pkgName} ({err})")

    return 0, None


def isOnSameFileSystem(fileName, path):
    isEqual = False
    try:
        isEqual = os.lstat(fileName).st_dev == os.lstat(os.path.split(path)[0]).st_dev
    except FileNotFoundError:
        pass

    return isEqual


def install(config, fileName, path, pkg):
    pkgName, _ = getPackageName(pkg)
    try:
        # move or copy/unlink data files from download dir to target path,
        # create softlink to for the download isDirty test
        if fileName.endswith((GRD_EXTENSION, MBTILES_EXTENSION)):
            if isOnSameFileSystem(fileName, path):
                progress = tqdm(
                    range(0, 1),
                    ascii=True,
                    desc=f"{log.getLogPrefix(INSTALL)}"
                    f"{pkgName: <{config.maxlenPackageFullName:}}",
                    bar_format=getTerminalSizeBarFormat(config),
                )
                # same file system: move
                for _ in progress:
                    os.replace(fileName, path)
                    if os.path.islink(fileName):
                        os.unlink(fileName)
                    os.symlink(path, fileName)
                    progress.set_description(
                        f"{log.getLogPrefix(INSTALL_OK)}"
                        f"{pkgName: <{config.maxlenPackageFullName:}}"
                    )
            # different filesystem: copy and unlink
            else:
                copy(config, fileName, path, pkgName)
                os.unlink(fileName)
                os.symlink(path, fileName)

            return True, 1, None

        # unzip
        if fileName.endswith(ZIP_EXTENSION):
            return True, unzip(config, fileName, path, pkg), None

        # untar
        fileCount, installFiles = untar(config, fileName, path, pkg)
        return True, fileCount, installFiles

    except Exception as ex:
        progress.set_description(
            f"{log.getLogPrefix(INSTALL_OK)}{pkgName: <{config.maxlenPackageFullName:}}"
        )
        log.logError(f"Failed to extract {pkgName} ({ex})")

    return False, 0, None


def cleanDirectory(directory, directoryName):
    if os.path.isdir(directory):
        try:
            usedDiskSpace = du(directory)
            print(f"{log.getLogPrefix('CLEAN')}{directoryName} ... ", end="")
            shutil.rmtree(directory)
            print(f"OK: disk space saved: ({directory}, {usedDiskSpace})")
        except OSError as ex:
            log.logError(str(ex))
            return False
    else:
        print(
            f"{log.getLogPrefix('ERROR')}{directoryName} does not exist ({directory})"
        )
    return True


def clean(patterns, config):
    if "packages" in patterns:
        cleanDirectory(config.downloadDir, "packages")

    if "dbcache" in patterns:
        cleanDirectory(config.syncDir, "dbcache")

    if "all" in patterns:
        cleanDirectory(config.syncDir, "dbcache")
        cleanDirectory(config.downloadDir, "packages")

    return True


def checkFilesExist(tarFile, installPath):
    try:
        with tarfile.open(tarFile, "r") as tar:
            for member in tar.getmembers():
                if os.path.isfile(os.path.join(installPath, member.name)):
                    return member.name
    except tarfile.ReadError:
        pass

    return None


def createDirectory(path):
    if os.path.exists(path):
        return True

    try:
        os.makedirs(path)
    except OSError as e:
        log.logError(f"Failed to create directory '{path}': {e}")
        return False

    return True


def detectSCPython(seiscompRoot, isSC3=False):
    soPath = os.path.join(
        seiscompRoot,
        "lib/python/seiscomp3/_Core.so" if isSC3 else "lib/python/seiscomp/_core.so",
    )
    if os.path.isfile(soPath):
        with Popen(["ldd", soPath], stdout=PIPE) as process:
            while True:
                line = process.stdout.readline()
                if not line:
                    break

                line = line.decode("utf-8", "replace")
                pos = line.find("libpython")
                if pos >= 0 and len(line) > pos + 9:
                    version = line[pos + 9]
                    return version

    raise Exception(f"Could not find libpython using ldd on {soPath}")


# -----------------------------------------------------------------------------
def createWrapper(config):
    template = """#!/bin/sh
#
# This is a shell script that executes the Python interpreter
# matching your SeisComP release as detected by
# gsm (gempa software manager). This file may be overridden by
# future SeisComP releases.
#
# In order to use this in your Python programs use this
# shebang line:

#!/usr/bin/env seiscomp-python

# Please note that this wrapper does *not* set the environment
# variables for you. To ensure that you run your script in the
# proper environment, please use 'seiscomp exec'. Alternatively
# you can also set your environment variables according to the
# output of 'seiscomp print env'.

python_executable="/usr/bin/python{}"

exec $python_executable "$@"
"""

    seiscompRoot = os.path.join(config.installDir, config.rootDir)
    wrapper = os.path.join(seiscompRoot, "bin", "seiscomp-python")
    isSC3 = config.isSC3()

    if os.path.isfile(wrapper):
        return True

    try:
        version = detectSCPython(seiscompRoot, isSC3)
    except Exception as e:
        log.logWarning(f"Could not determine SeisComP Python version: {e}")
        return False

    try:
        with open(wrapper, "w", encoding="utf-8") as f:
            f.write(template.format(version))
            os.chmod(wrapper, 0o775)
    except Exception as e:
        log.logError(f"Could not generate wrapper {wrapper}: {e}")
        return False

    return True


def getLicenseTimeDelta(cert):
    try:
        updatesUntilStr = str(cert.issuer).split("updatesUntil=")[1].split(" ")[0]
    except IndexError:
        updatesUntilStr = "9999-12-31"

    notValidAfterStr = cert.not_valid_after.strftime(DEFAULT_TIME_FORMAT)
    todayStr = date.today().strftime(DEFAULT_TIME_FORMAT)

    notValidAfter = datetime.strptime(notValidAfterStr, DEFAULT_TIME_FORMAT)
    updatesUntil = datetime.strptime(updatesUntilStr, DEFAULT_TIME_FORMAT)
    today = datetime.strptime(todayStr, DEFAULT_TIME_FORMAT)

    deltaUpdatesUntil = int((updatesUntil - today).days)
    deltaNotValidAfter = int((notValidAfter - today).days)

    return (deltaUpdatesUntil, deltaNotValidAfter, updatesUntilStr, notValidAfterStr)


def prettyFormatLicenseInfo(cert, licenseInfoDays):
    licenseInfoDays = int(licenseInfoDays)
    (
        deltaUpdatesUntil,
        deltaNotValidAfter,
        updatesUntil,
        notValidAfter,
    ) = getLicenseTimeDelta(cert)

    if licenseInfoDays <= deltaUpdatesUntil:
        deltaUpdatesUntil = None

    if licenseInfoDays <= deltaNotValidAfter:
        deltaNotValidAfter = None

    lincenseInfo = ""

    def helper(days, infoType, infoDate):

        if infoType == "valid":
            if days <= 0:
                return f"License expired on {infoDate}"
            return f"License expires in {days} days ({infoDate})"
        if days <= 0:
            return f"Update period expired on {infoDate}"

        return f"Update period ends in {days} days ({infoDate})"

    if deltaUpdatesUntil and deltaNotValidAfter:
        if deltaUpdatesUntil == deltaNotValidAfter:
            lincenseInfo = f"[{helper(deltaNotValidAfter, 'valid', notValidAfter)}]"
        else:
            lincenseInfo = (
                f"[{helper(deltaNotValidAfter,'valid', notValidAfter)}, "
                f"{helper(deltaUpdatesUntil, 'license', updatesUntil)}]"
            )

    if deltaUpdatesUntil and not deltaNotValidAfter:
        lincenseInfo = f"[{helper(deltaUpdatesUntil,'license', updatesUntil)}]"

    if not deltaUpdatesUntil and deltaNotValidAfter:
        lincenseInfo = f"([{helper(deltaNotValidAfter,'valid', notValidAfter )}]"

    return lincenseInfo


def getDownloadDirectoryURL(name, pkg, config):
    if pkg.type == PT.CONFIG:
        url = urljoin(config.baseURL, pkg.name)
    elif pkg.filePath:
        url = urljoin(
            config.baseURL, os.path.join(os.path.split(pkg.filePath)[0], pkg.name)
        )
    elif name in PUBLIC_PACKAGES:
        url = urljoin(BASE_PUBLIC_URL, name)
    else:
        url = urljoin(config.baseURL, name)
    return url


def splitVersionString(version):
    if version == "VERSION":
        version = "9.9.9"
    return tuple(map(int, (version.split("."))))


# tries to calculate the length of the package names and get a constant distance
# so that the progressbars start at the same place


def setMaxlenPackageFullName(packages, localCache, syncCache, config):
    packages = list(set(list(localCache.getIndex().keys()) + packages))

    headerPackages = ""
    packagesFullName = []
    downloadLicense = []
    try:
        for name in packages:
            pkg = syncCache.getPackages(name)[0]
            pkgName, _ = getPackageName(pkg)
            packagesFullName.append(pkgName)
    except IndexError:
        packagesFullName = []

    if len(packagesFullName) == 0:
        config.maxlenPackageFullName = MAX_LEN_PACKAGE_FULL_NAME
    else:
        maxlenPackageFullName = len(max(packagesFullName, key=len))
        maxlenPackageFullName = min(
            max(maxlenPackageFullName, len(headerPackages)), MAX_LEN_PACKAGE_FULL_NAME
        )

        for name in packages:
            pkg = syncCache.getPackages(name)[0]

            if not localCache.isInstalled(pkg, False):
                fileName = os.path.join(
                    config.downloadLicenseDir, pkg.name + LICENSE_TXT_EXTENSION
                )
                if pkg.license == "Gempa" and not os.path.exists(fileName):
                    downloadLicense.append(pkg.name)

        minLenChangelog = len(sorting(packages)[-1])
        if maxlenPackageFullName < (minLenChangelog + len(f" {CHANGELOG} ")):
            maxlenPackageFullNameChangelog = minLenChangelog + len(f" {CHANGELOG} ")
        else:
            maxlenPackageFullNameChangelog = maxlenPackageFullName

        if downloadLicense:
            minLenLicense = len(sorting(downloadLicense)[-1])
            if maxlenPackageFullName < (
                minLenLicense + len(f"{LICENSE_TXT_EXTENSION} ")
            ):
                maxlenPackageFullNameLicense = minLenLicense + len(
                    f"{LICENSE_TXT_EXTENSION} "
                )
            else:
                maxlenPackageFullNameLicense = maxlenPackageFullName
        else:
            maxlenPackageFullNameLicense = maxlenPackageFullName

        if maxlenPackageFullNameChangelog > maxlenPackageFullNameLicense:
            config.maxlenPackageFullName = maxlenPackageFullNameChangelog
        else:
            config.maxlenPackageFullName = maxlenPackageFullNameLicense
